import pytest

import labelbox as lb
import labelbox.types as lb_types
from labelbox.data.annotation_types.data import GenericDataRowData
from labelbox.schema.annotation_import import AnnotationImportState
from labelbox.schema.export_task import ExportTask, StreamType


class TestExportVideo:
    @pytest.fixture
    def user_id(self, client):
        return client.get_user().uid

    @pytest.fixture
    def org_id(self, client):
        return client.get_organization().uid

    def test_export(
        self,
        client,
        configured_video_project_without_data_rows,
        video_data,
        video_data_row,
        bbox_video_annotation_objects,
        rand_gen,
    ):
        project = configured_video_project_without_data_rows
        project_id = project.uid
        labels = []

        _, data_row_uids = video_data
        project.create_batch(
            rand_gen(str),
            data_row_uids,  # sample of data row objects
            5,  # priority between 1(Highest) - 5(lowest)
        )

        for data_row_uid in data_row_uids:
            labels = [
                lb_types.Label(
                    data=GenericDataRowData(uid=data_row_uid),
                    annotations=bbox_video_annotation_objects,
                )
            ]

        label_import = lb.LabelImport.create_from_objects(
            client, project_id, f"test-import-{project_id}", labels
        )
        label_import.wait_until_done()

        assert label_import.state == AnnotationImportState.FINISHED
        assert len(label_import.errors) == 0

        export_task = project.export(
            params={
                "performance_details": False,
                "label_details": True,
                "interpolated_frames": True,
            },
            task_name="TestExportVideo:test_export",
        )
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert isinstance(export_task, ExportTask)
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )

        export_data = list(export_task.get_buffered_stream())[0].json

        data_row_export = export_data["data_row"]
        assert data_row_export["global_key"] == video_data_row["global_key"]
        assert data_row_export["row_data"] == video_data_row["row_data"]
        assert export_data["media_attributes"]["mime_type"] == "video/mp4"
        assert (
            export_data["media_attributes"]["frame_rate"] == 10
        )  # as per the video_data fixture
        assert (
            export_data["media_attributes"]["frame_count"] == 100
        )  # as per the video_data fixture
        expected_export_label = {
            "label_kind": "Video",
            "version": "1.0.0",
            "id": "clgjnpysl000xi3zxtnp29fug",
            "label_details": {
                "created_at": "2023-04-16T17:04:23+00:00",
                "updated_at": "2023-04-16T17:04:23+00:00",
                "created_by": "vbrodsky@labelbox.com",
                "content_last_updated_at": "2023-04-16T17:04:23+00:00",
                "reviews": [],
            },
            "annotations": {
                "frames": {
                    "13": {
                        "objects": {
                            "clgjnpyse000ui3zx6fr1d880": {
                                "feature_id": "clgjnpyse000ui3zx6fr1d880",
                                "name": "bbox",
                                "annotation_kind": "VideoBoundingBox",
                                "classifications": [
                                    {
                                        "feature_id": "clgjnpyse000vi3zxtgtfh01y",
                                        "name": "nested",
                                        "radio_answer": {
                                            "feature_id": "clgjnpyse000wi3zxnxgv53ps",
                                            "name": "radio_option_1",
                                            "classifications": [],
                                        },
                                    }
                                ],
                                "bounding_box": {
                                    "top": 98.0,
                                    "left": 146.0,
                                    "height": 243.0,
                                    "width": 236.0,
                                },
                            }
                        },
                        "classifications": [],
                    },
                    "18": {
                        "objects": {
                            "clgjnpyse000ui3zx6fr1d880": {
                                "feature_id": "clgjnpyse000ui3zx6fr1d880",
                                "name": "bbox",
                                "annotation_kind": "VideoBoundingBox",
                                "classifications": [
                                    {
                                        "feature_id": "clgjnpyse000vi3zxtgtfh01y",
                                        "name": "nested",
                                        "radio_answer": {
                                            "feature_id": "clgjnpyse000wi3zxnxgv53ps",
                                            "name": "radio_option_1",
                                            "classifications": [],
                                        },
                                    }
                                ],
                                "bounding_box": {
                                    "top": 98.0,
                                    "left": 146.0,
                                    "height": 243.0,
                                    "width": 236.0,
                                },
                            }
                        },
                        "classifications": [],
                    },
                    "19": {
                        "objects": {
                            "clgjnpyse000ui3zx6fr1d880": {
                                "feature_id": "clgjnpyse000ui3zx6fr1d880",
                                "name": "bbox",
                                "annotation_kind": "VideoBoundingBox",
                                "classifications": [],
                                "bounding_box": {
                                    "top": 98.0,
                                    "left": 146.0,
                                    "height": 243.0,
                                    "width": 236.0,
                                },
                            }
                        },
                        "classifications": [],
                    },
                },
                "segments": {"clgjnpyse000ui3zx6fr1d880": [[13, 13], [18, 19]]},
                "key_frame_feature_map": {
                    "clgjnpyse000ui3zx6fr1d880": {
                        "13": True,
                        "18": False,
                        "19": True,
                    }
                },
                "classifications": [],
            },
        }

        project_export_labels = export_data["projects"][project_id]["labels"]
        assert len(project_export_labels) == len(
            labels
        )  # note we create 1 label per data row, 1 data row so 1 label
        export_label = project_export_labels[0]
        assert (export_label["label_kind"]) == "Video"

        assert (export_label["label_details"].keys()) == expected_export_label[
            "label_details"
        ].keys()

        expected_frames_ids = [
            vannotation.frame for vannotation in bbox_video_annotation_objects
        ]
        export_annotations = export_label["annotations"]
        export_frames = export_annotations["frames"]
        export_frames_ids = [int(frame_id) for frame_id in export_frames.keys()]
        all_frames_exported = []
        for value in expected_frames_ids:  # note need to understand why we are exporting more frames than we created
            if value not in export_frames_ids:
                all_frames_exported.append(value)
        assert len(all_frames_exported) == 0

        # BEGINNING OF THE VIDEO INTERPOLATION ASSERTIONS
        first_frame_id = bbox_video_annotation_objects[0].frame
        last_frame_id = bbox_video_annotation_objects[-1].frame

        # Generate list of frames with frames in between, e.g. 13, 14, 15, 16, 17, 18, 19
        expected_frame_ids = list(range(first_frame_id, last_frame_id + 1))

        assert export_frames_ids == expected_frame_ids

        exported_objects_dict = export_frames[str(first_frame_id)]["objects"]

        # Get the label ID
        first_exported_label_id = list(exported_objects_dict.keys())[0]

        # Since the bounding box moves to the right, the interpolated frame content should start
        # a little bit more far to the right
        assert (
            export_frames[str(first_frame_id + 1)]["objects"][
                first_exported_label_id
            ]["bounding_box"]["left"]
            > export_frames[str(first_frame_id)]["objects"][
                first_exported_label_id
            ]["bounding_box"]["left"]
        )
        # But it shouldn't be further than the last frame
        assert (
            export_frames[str(first_frame_id + 1)]["objects"][
                first_exported_label_id
            ]["bounding_box"]["left"]
            < export_frames[str(last_frame_id)]["objects"][
                first_exported_label_id
            ]["bounding_box"]["left"]
        )
        # END OF THE VIDEO INTERPOLATION ASSERTIONS

        frame_with_nested_classifications = export_frames["13"]
        annotation = None
        for _, a in frame_with_nested_classifications["objects"].items():
            if a["name"] == "bbox":
                annotation = a
                break
        assert annotation is not None
        assert annotation["annotation_kind"] == "VideoBoundingBox"
        assert annotation["classifications"]
        assert annotation["bounding_box"] == {
            "top": 98.0,
            "left": 146.0,
            "height": 243.0,
            "width": 236.0,
        }
        classifications = annotation["classifications"]
        classification = classifications[0]["radio_answer"]
        assert classification["name"] == "radio_option_1"
        subclassifications = classification["classifications"]
        # NOTE predictions services does not support nested classifications at the moment, see
        # https://labelbox.atlassian.net/browse/AL-5588
        assert len(subclassifications) == 0



import datetime
import itertools
import uuid

import pytest

import labelbox as lb
from labelbox import Client, OntologyKind, Project
from labelbox.schema.annotation_import import AnnotationImportState
from labelbox.schema.media_type import MediaType

"""
 - integration test for importing mal labels and ground truths with each supported MediaType. 
 - NDJSON is used to generate annotations.
"""


def validate_iso_format(date_string: str):
    parsed_t = datetime.datetime.fromisoformat(
        date_string
    )  # this will blow up if the string is not in iso format
    assert parsed_t.hour is not None
    assert parsed_t.minute is not None
    assert parsed_t.second is not None


@pytest.mark.parametrize(
    "configured_project, media_type",
    [
        (MediaType.Audio, MediaType.Audio),
        (MediaType.Html, MediaType.Html),
        (MediaType.Image, MediaType.Image),
        (MediaType.Text, MediaType.Text),
        (MediaType.Video, MediaType.Video),
        (MediaType.Conversational, MediaType.Conversational),
        (MediaType.Document, MediaType.Document),
        (OntologyKind.ResponseCreation, OntologyKind.ResponseCreation),
        (OntologyKind.ModelEvaluation, OntologyKind.ModelEvaluation),
    ],
    indirect=["configured_project"],
)
def test_import_media_types(
    client: Client,
    configured_project: Project,
    annotations_by_media_type,
    exports_v2_by_media_type,
    export_v2_test_helpers,
    helpers,
    media_type,
    wait_for_label_processing,
):
    annotations_ndjson = list(
        itertools.chain.from_iterable(annotations_by_media_type[media_type])
    )

    label_import = lb.LabelImport.create_from_objects(
        client,
        configured_project.uid,
        f"test-import-{media_type}",
        annotations_ndjson,
    )
    label_import.wait_until_done()

    assert label_import.state == AnnotationImportState.FINISHED
    assert len(label_import.errors) == 0

    wait_for_label_processing(configured_project)[0]

    result = export_v2_test_helpers.run_project_export_v2_task(
        configured_project
    )

    assert result

    for exported_data in result:
        # timestamp fields are in iso format
        validate_iso_format(exported_data["data_row"]["details"]["created_at"])
        validate_iso_format(exported_data["data_row"]["details"]["updated_at"])
        validate_iso_format(
            exported_data["projects"][configured_project.uid]["labels"][0][
                "label_details"
            ]["created_at"]
        )
        validate_iso_format(
            exported_data["projects"][configured_project.uid]["labels"][0][
                "label_details"
            ]["updated_at"]
        )

        assert (
            exported_data["data_row"]["id"] in configured_project.data_row_ids
        )
        exported_project = exported_data["projects"][configured_project.uid]
        exported_project_labels = exported_project["labels"][0]
        exported_annotations = exported_project_labels["annotations"]

        expected_data = exports_v2_by_media_type[media_type]
        helpers.remove_keys_recursive(
            exported_annotations, ["feature_id", "feature_schema_id"]
        )
        helpers.rename_cuid_key_recursive(exported_annotations)

        assert exported_annotations == expected_data


@pytest.mark.parametrize(
    "configured_project, media_type",
    [
        (
            MediaType.LLMPromptResponseCreation,
            MediaType.LLMPromptResponseCreation,
        ),
        (MediaType.LLMPromptCreation, MediaType.LLMPromptCreation),
    ],
    indirect=["configured_project"],
)
def test_import_media_types_llm(
    client: Client,
    configured_project: Project,
    annotations_by_media_type,
    exports_v2_by_media_type,
    export_v2_test_helpers,
    helpers,
    media_type,
    wait_for_label_processing,
):
    annotations_ndjson = list(
        itertools.chain.from_iterable(annotations_by_media_type[media_type])
    )

    label_import = lb.LabelImport.create_from_objects(
        client,
        configured_project.uid,
        f"test-import-{media_type}",
        annotations_ndjson,
    )
    label_import.wait_until_done()

    assert label_import.state == AnnotationImportState.FINISHED
    assert len(label_import.errors) == 0

    all_annotations = sorted([a["uuid"] for a in annotations_ndjson])
    successful_annotations = sorted(
        [
            status["uuid"]
            for status in label_import.statuses
            if status["status"] == "SUCCESS"
        ]
    )
    assert successful_annotations == all_annotations


@pytest.mark.parametrize(
    "configured_project_by_global_key, media_type",
    [
        (MediaType.Audio, MediaType.Audio),
        (MediaType.Html, MediaType.Html),
        (MediaType.Image, MediaType.Image),
        (MediaType.Text, MediaType.Text),
        (MediaType.Video, MediaType.Video),
        (MediaType.Conversational, MediaType.Conversational),
        (MediaType.Document, MediaType.Document),
        (OntologyKind.ResponseCreation, OntologyKind.ResponseCreation),
        (OntologyKind.ModelEvaluation, OntologyKind.ModelEvaluation),
    ],
    indirect=["configured_project_by_global_key"],
)
def test_import_media_types_by_global_key(
    client,
    configured_project_by_global_key,
    annotations_by_media_type,
    exports_v2_by_media_type,
    export_v2_test_helpers,
    helpers,
    media_type,
):
    annotations_ndjson = list(
        itertools.chain.from_iterable(annotations_by_media_type[media_type])
    )

    label_import = lb.LabelImport.create_from_objects(
        client,
        configured_project_by_global_key.uid,
        f"test-import-{media_type}",
        annotations_ndjson,
    )
    label_import.wait_until_done()

    assert label_import.state == AnnotationImportState.FINISHED
    assert len(label_import.errors) == 0

    result = export_v2_test_helpers.run_project_export_v2_task(
        configured_project_by_global_key
    )

    assert result

    for exported_data in result:
        # timestamp fields are in iso format
        validate_iso_format(exported_data["data_row"]["details"]["created_at"])
        validate_iso_format(exported_data["data_row"]["details"]["updated_at"])
        validate_iso_format(
            exported_data["projects"][configured_project_by_global_key.uid][
                "labels"
            ][0]["label_details"]["created_at"]
        )
        validate_iso_format(
            exported_data["projects"][configured_project_by_global_key.uid][
                "labels"
            ][0]["label_details"]["updated_at"]
        )

        assert (
            exported_data["data_row"]["id"]
            in configured_project_by_global_key.data_row_ids
        )
        exported_project = exported_data["projects"][
            configured_project_by_global_key.uid
        ]
        exported_project_labels = exported_project["labels"][0]
        exported_annotations = exported_project_labels["annotations"]

        expected_data = exports_v2_by_media_type[media_type]
        helpers.remove_keys_recursive(
            exported_annotations, ["feature_id", "feature_schema_id"]
        )
        helpers.rename_cuid_key_recursive(exported_annotations)

        assert exported_annotations == expected_data


@pytest.mark.parametrize(
    "configured_project, media_type",
    [
        (MediaType.Audio, MediaType.Audio),
        (MediaType.Html, MediaType.Html),
        (MediaType.Image, MediaType.Image),
        (MediaType.Text, MediaType.Text),
        (MediaType.Video, MediaType.Video),
        (MediaType.Conversational, MediaType.Conversational),
        (MediaType.Document, MediaType.Document),
        (
            MediaType.LLMPromptResponseCreation,
            MediaType.LLMPromptResponseCreation,
        ),
        (MediaType.LLMPromptCreation, MediaType.LLMPromptCreation),
        (OntologyKind.ResponseCreation, OntologyKind.ResponseCreation),
        (OntologyKind.ModelEvaluation, OntologyKind.ModelEvaluation),
    ],
    indirect=["configured_project"],
)
def test_import_mal_annotations(
    client, configured_project: Project, annotations_by_media_type, media_type
):
    annotations_ndjson = list(
        itertools.chain.from_iterable(annotations_by_media_type[media_type])
    )

    import_annotations = lb.MALPredictionImport.create_from_objects(
        client=client,
        project_id=configured_project.uid,
        name=f"import {str(uuid.uuid4())}",
        predictions=annotations_ndjson,
    )
    import_annotations.wait_until_done()

    assert import_annotations.errors == []
    # MAL Labels cannot be exported and compared to input labels


@pytest.mark.parametrize(
    "configured_project_by_global_key, media_type",
    [
        (MediaType.Audio, MediaType.Audio),
        (MediaType.Html, MediaType.Html),
        (MediaType.Image, MediaType.Image),
        (MediaType.Text, MediaType.Text),
        (MediaType.Video, MediaType.Video),
        (MediaType.Conversational, MediaType.Conversational),
        (MediaType.Document, MediaType.Document),
        (OntologyKind.ResponseCreation, OntologyKind.ResponseCreation),
        (OntologyKind.ModelEvaluation, OntologyKind.ModelEvaluation),
    ],
    indirect=["configured_project_by_global_key"],
)
def test_import_mal_annotations_global_key(
    client,
    configured_project_by_global_key: Project,
    annotations_by_media_type,
    media_type,
):
    annotations_ndjson = list(
        itertools.chain.from_iterable(annotations_by_media_type[media_type])
    )

    import_annotations = lb.MALPredictionImport.create_from_objects(
        client=client,
        project_id=configured_project_by_global_key.uid,
        name=f"import {str(uuid.uuid4())}",
        predictions=annotations_ndjson,
    )
    import_annotations.wait_until_done()

    assert import_annotations.errors == []
    # MAL Labels cannot be exported and compared to input labels



from typing import Callable, Literal, Optional

from labelbox.data.annotation_types.data.base_data import BaseData
from labelbox.utils import _NoCoercionMixin
from pydantic import model_validator


class GenericDataRowData(BaseData, _NoCoercionMixin):
    """Generic data row data. This is replacing all other DataType passed into Label"""

    url: Optional[str] = None
    class_name: Literal["GenericDataRowData"] = "GenericDataRowData"

    def create_url(self, signer: Callable[[bytes], str]) -> Optional[str]:
        return self.url

    @model_validator(mode="before")
    @classmethod
    def validate_one_datarow_key_present(cls, data):
        keys = ["external_id", "global_key", "uid"]
        count = sum([key in data for key in keys])

        if count < 1:
            raise ValueError(f"Exactly one of {keys} must be present.")
        if count > 1:
            raise ValueError(f"Only one of {keys} can be present.")
        return data



from dataclasses import dataclass, field
from enum import Enum

from labelbox.schema.tool_building.base_step_reasoning_tool import (
    _BaseStepReasoningTool,
    _Definition,
    _Variant,
)
from labelbox.schema.tool_building.tool_type import ToolType


class FactCheckingActions(Enum):
    WRITE_JUSTIFICATION = "justification"


def build_fact_checking_definition():
    accurate_step = _Variant(
        id=0,
        name="Accurate",
        _available_actions={action.value for action in FactCheckingActions},
        actions=[action.value for action in FactCheckingActions],
    )
    inaccurate_step = _Variant(
        id=1,
        name="Inaccurate",
        _available_actions={action.value for action in FactCheckingActions},
        actions=[action.value for action in FactCheckingActions],
    )
    disputed_step = _Variant(
        id=2,
        name="Disputed",
        _available_actions={action.value for action in FactCheckingActions},
        actions=[action.value for action in FactCheckingActions],
    )
    unsupported_step = _Variant(
        id=3,
        name="Unsupported",
        _available_actions=set(),
        actions=[],
    )
    cant_confidently_assess_step = _Variant(
        id=4,
        name="Can't confidently assess",
        _available_actions=set(),
        actions=[],
    )
    no_factual_information_step = _Variant(
        id=5,
        name="No factual information",
        _available_actions=set(),
        actions=[],
    )
    variants = [
        accurate_step,
        inaccurate_step,
        disputed_step,
        unsupported_step,
        cant_confidently_assess_step,
        no_factual_information_step,
    ]
    return _Definition(variants=variants)


@dataclass
class FactCheckingTool(_BaseStepReasoningTool):
    """
    Use this class in OntologyBuilder to create a tool for fact checking

    Note variant kinds can not be changed
    """

    type: ToolType = field(default=ToolType.FACT_CHECKING, init=False)
    definition: _Definition = field(
        default_factory=build_fact_checking_definition
    )

    def __post_init__(self):
        super().__post_init__()
        # Set available actions for variants 0, 1, 2 'out of band' since they are not passed in the definition
        self._set_variant_available_actions()

    def _set_variant_available_actions(self):
        for variant in self.definition.variants:
            if variant.id in [0, 1, 2]:
                for action in FactCheckingActions:
                    variant._available_actions.add(action.value)



import sys

if sys.version_info >= (3, 9):
    from strenum import StrEnum

    class BaseStrEnum(StrEnum):
        pass
else:
    from enum import Enum

    class BaseStrEnum(str, Enum):
        pass


class IdType(BaseStrEnum):
    """
    The type of id used to identify a data row.

    Currently supported types are:
        - DataRowId: The id assigned to a data row by Labelbox.
        - GlobalKey: The id assigned to a data row by the user.
    """

    DataRowId = "ID"
    GlobalKey = "GKEY"



import json
import os
import re
import time
import uuid
from datetime import datetime
from enum import Enum
from random import randint
from string import ascii_letters
from types import SimpleNamespace
from typing import Tuple, Type

import pytest
import requests
from lbox.exceptions import LabelboxError

from labelbox import (
    Classification,
    Client,
    DataRow,
    Dataset,
    MediaType,
    OntologyBuilder,
    Option,
    Tool,
)
from labelbox.orm import query
from labelbox.pagination import PaginatedCollection
from labelbox.schema.annotation_import import LabelImport
from labelbox.schema.enums import AnnotationImportState
from labelbox.schema.invite import Invite
from labelbox.schema.ontology import Ontology
from labelbox.schema.project import Project
from labelbox.schema.quality_mode import QualityMode

IMG_URL = "https://picsum.photos/200/300.jpg"
MASKABLE_IMG_URL = "https://storage.googleapis.com/labelbox-datasets/image_sample_data/2560px-Kitano_Street_Kobe01s5s4110.jpeg"
SMALL_DATASET_URL = "https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/potato.jpeg"
DATA_ROW_PROCESSING_WAIT_TIMEOUT_SECONDS = 30
DATA_ROW_PROCESSING_WAIT_SLEEP_INTERNAL_SECONDS = 3
EPHEMERAL_BASE_URL = "http://lb-api-public"
IMAGE_URL = "https://storage.googleapis.com/diagnostics-demo-data/coco/COCO_train2014_000000000034.jpg"
EXTERNAL_ID = "my-image"

pytest_plugins = []


@pytest.fixture(scope="session")
def rand_gen():
    def gen(field_type):
        if field_type is str:
            return "".join(
                ascii_letters[randint(0, len(ascii_letters) - 1)]
                for _ in range(16)
            )

        if field_type is datetime:
            return datetime.now()

        raise Exception(
            "Can't random generate for field type '%r'" % field_type
        )

    return gen


class Environ(Enum):
    LOCAL = "local"
    PROD = "prod"
    STAGING = "staging"
    CUSTOM = "custom"
    STAGING_EU = "staging-eu"
    EPHEMERAL = "ephemeral"  # Used for testing PRs with ephemeral environments


@pytest.fixture
def image_url() -> str:
    return MASKABLE_IMG_URL


@pytest.fixture
def external_id() -> str:
    return EXTERNAL_ID


def ephemeral_endpoint() -> str:
    return os.getenv("LABELBOX_TEST_BASE_URL", EPHEMERAL_BASE_URL)


def graphql_url(environ: str) -> str:
    if environ == Environ.LOCAL:
        return "http://localhost:3000/api/graphql"
    elif environ == Environ.PROD:
        return "https://api.labelbox.com/graphql"
    elif environ == Environ.STAGING:
        return "https://api.lb-stage.xyz/graphql"
    elif environ == Environ.CUSTOM:
        graphql_api_endpoint = os.environ.get(
            "LABELBOX_TEST_GRAPHQL_API_ENDPOINT"
        )
        if graphql_api_endpoint is None:
            raise Exception("Missing LABELBOX_TEST_GRAPHQL_API_ENDPOINT")
        return graphql_api_endpoint
    elif environ == Environ.EPHEMERAL:
        return f"{ephemeral_endpoint()}/graphql"
    return "http://host.docker.internal:8080/graphql"


def rest_url(environ: str) -> str:
    if environ == Environ.LOCAL:
        return "http://localhost:3000/api/v1"
    elif environ == Environ.PROD:
        return "https://api.labelbox.com/api/v1"
    elif environ == Environ.STAGING:
        return "https://api.lb-stage.xyz/api/v1"
    elif environ == Environ.CUSTOM:
        rest_api_endpoint = os.environ.get("LABELBOX_TEST_REST_API_ENDPOINT")
        if rest_api_endpoint is None:
            raise Exception("Missing LABELBOX_TEST_REST_API_ENDPOINT")
        return rest_api_endpoint
    elif environ == Environ.EPHEMERAL:
        return f"{ephemeral_endpoint()}/api/v1"
    return "http://host.docker.internal:8080/api/v1"


def testing_api_key(environ: Environ) -> str:
    keys = [
        f"LABELBOX_TEST_API_KEY_{environ.value.upper()}",
        "LABELBOX_TEST_API_KEY",
        "LABELBOX_API_KEY",
    ]
    for key in keys:
        value = os.environ.get(key)
        if value is not None:
            return value
    raise Exception("Cannot find API to use for tests")


def service_api_key() -> str:
    service_api_key = os.environ["SERVICE_API_KEY"]
    if service_api_key is None:
        raise Exception(
            "SERVICE_API_KEY is missing and needed for admin client"
        )
    return service_api_key


class IntegrationClient(Client):
    def __init__(self, environ: str) -> None:
        api_url = graphql_url(environ)
        api_key = testing_api_key(environ)
        rest_endpoint = rest_url(environ)
        super().__init__(
            api_key,
            api_url,
            enable_experimental=True,
            rest_endpoint=rest_endpoint,
        )
        self.queries = []

    def execute(self, query=None, params=None, check_naming=True, **kwargs):
        if check_naming and query is not None:
            assert (
                re.match(r"\s*(?:query|mutation) \w+PyApi", query) is not None
            )
        self.queries.append((query, params))
        if not kwargs.get("timeout"):
            kwargs["timeout"] = 30.0
        return super().execute(query, params, **kwargs)


class AdminClient(Client):
    def __init__(self, env):
        """
        The admin client creates organizations and users using admin api described here https://labelbox.atlassian.net/wiki/spaces/AP/pages/2206564433/Internal+Admin+APIs.
        """
        self._api_key = service_api_key()
        self._admin_endpoint = f"{ephemeral_endpoint()}/admin/v1"
        self._api_url = graphql_url(env)
        self._rest_endpoint = rest_url(env)

        super().__init__(
            self._api_key,
            self._api_url,
            enable_experimental=True,
            rest_endpoint=self._rest_endpoint,
        )

    def _create_organization(self) -> str:
        endpoint = f"{self._admin_endpoint}/organizations/"
        response = requests.post(
            endpoint,
            headers=self.headers,
            json={"name": f"Test Org {uuid.uuid4()}"},
        )

        data = response.json()
        if response.status_code not in [
            requests.codes.created,
            requests.codes.ok,
        ]:
            raise Exception(
                "Failed to create org, message: " + str(data["message"])
            )

        return data["id"]

    def _create_user(self, organization_id=None) -> Tuple[str, str]:
        if organization_id is None:
            organization_id = self.organization_id

        endpoint = f"{self._admin_endpoint}/user-identities/"
        identity_id = f"e2e+{uuid.uuid4()}"

        response = requests.post(
            endpoint,
            headers=self.headers,
            json={
                "identityId": identity_id,
                "email": "email@email.com",
                "name": f"tester{uuid.uuid4()}",
                "verificationStatus": "VERIFIED",
            },
        )
        data = response.json()
        if response.status_code not in [
            requests.codes.created,
            requests.codes.ok,
        ]:
            raise Exception(
                "Failed to create user, message: " + str(data["message"])
            )

        user_identity_id = data["identityId"]

        endpoint = (
            f"{self._admin_endpoint}/organizations/{organization_id}/users/"
        )
        response = requests.post(
            endpoint,
            headers=self.headers,
            json={"identityId": user_identity_id, "organizationRole": "Admin"},
        )

        data = response.json()
        if response.status_code not in [
            requests.codes.created,
            requests.codes.ok,
        ]:
            raise Exception(
                "Failed to create link user to org, message: "
                + str(data["message"])
            )

        user_id = data["id"]

        endpoint = f"{self._admin_endpoint}/users/{user_id}/token"
        response = requests.get(
            endpoint,
            headers=self.headers,
        )
        data = response.json()
        if response.status_code not in [
            requests.codes.created,
            requests.codes.ok,
        ]:
            raise Exception(
                "Failed to create ephemeral user, message: "
                + str(data["message"])
            )

        token = data["token"]

        return user_id, token

    def create_api_key_for_user(self) -> str:
        organization_id = self._create_organization()
        _, user_token = self._create_user(organization_id)
        key_name = f"test-key+{uuid.uuid4()}"
        query = """mutation CreateApiKeyPyApi($name: String!) {
                createApiKey(data: {name: $name}) {
                    id
                    jwt
                }
            }
        """
        params = {"name": key_name}
        self.headers["Authorization"] = f"Bearer {user_token}"
        res = self.execute(query, params, error_log_key="errors")

        return res["createApiKey"]["jwt"]


class EphemeralClient(Client):
    def __init__(self, environ=Environ.EPHEMERAL):
        self.admin_client = AdminClient(environ)
        self.api_key = self.admin_client.create_api_key_for_user()
        api_url = graphql_url(environ)
        rest_endpoint = rest_url(environ)

        super().__init__(
            self.api_key,
            api_url,
            enable_experimental=True,
            rest_endpoint=rest_endpoint,
        )


@pytest.fixture
def ephmeral_client() -> EphemeralClient:
    return EphemeralClient


@pytest.fixture
def admin_client() -> AdminClient:
    return AdminClient


@pytest.fixture
def integration_client() -> IntegrationClient:
    return IntegrationClient


@pytest.fixture(scope="session")
def environ() -> Environ:
    """
    Checks environment variables for LABELBOX_ENVIRON to be
    'prod' or 'staging'
    Make sure to set LABELBOX_TEST_ENVIRON in .github/workflows/python-package.yaml
    """
    keys = ["LABELBOX_TEST_ENV", "LABELBOX_TEST_ENVIRON", "LABELBOX_ENV"]
    for key in keys:
        value = os.environ.get(key)
        if value is not None:
            return Environ(value)
    raise Exception(f"Missing env key in: {os.environ}")


def cancel_invite(client, invite_id):
    """
    Do not use. Only for testing.
    """
    query_str = """mutation CancelInvitePyApi($where: WhereUniqueIdInput!) {
            cancelInvite(where: $where) {id}}"""
    client.execute(query_str, {"where": {"id": invite_id}}, experimental=True)


def get_project_invites(client, project_id):
    """
    Do not use. Only for testing.
    """
    id_param = "projectId"
    query_str = """query GetProjectInvitationsPyApi($from: ID, $first: PageSize, $%s: ID!) {
        project(where: {id: $%s}) {id
        invites(from: $from, first: $first) { nodes { %s
        projectInvites { projectId projectRoleName } } nextCursor}}}
    """ % (id_param, id_param, query.results_query_part(Invite))
    return PaginatedCollection(
        client,
        query_str,
        {id_param: project_id},
        ["project", "invites", "nodes"],
        Invite,
        cursor_path=["project", "invites", "nextCursor"],
    )


def get_invites(client):
    """
    Do not use. Only for testing.
    """
    query_str = """query GetOrgInvitationsPyApi($from: ID, $first: PageSize) {
            organization { id invites(from: $from, first: $first) {
                nodes { id createdAt organizationRoleName inviteeEmail } nextCursor }}}"""
    invites = PaginatedCollection(
        client,
        query_str,
        {},
        ["organization", "invites", "nodes"],
        Invite,
        cursor_path=["organization", "invites", "nextCursor"],
        experimental=True,
    )
    return invites


@pytest.fixture
def queries():
    return SimpleNamespace(
        cancel_invite=cancel_invite,
        get_project_invites=get_project_invites,
        get_invites=get_invites,
    )


@pytest.fixture(scope="session")
def admin_client(environ: str):
    return AdminClient(environ)


@pytest.fixture(scope="session")
def client(environ: str):
    if environ == Environ.EPHEMERAL:
        return EphemeralClient()
    return IntegrationClient(environ)


@pytest.fixture(scope="session")
def pdf_url(client):
    pdf_url = client.upload_file("tests/assets/loremipsum.pdf")
    return {
        "row_data": {
            "pdf_url": pdf_url,
        },
        "global_key": str(uuid.uuid4()),
    }


@pytest.fixture(scope="session")
def pdf_entity_data_row(client):
    pdf_url = client.upload_file(
        "tests/assets/arxiv-pdf_data_99-word-token-pdfs_0801.3483.pdf"
    )
    text_layer_url = client.upload_file(
        "tests/assets/arxiv-pdf_data_99-word-token-pdfs_0801.3483-lb-textlayer.json"
    )

    return {
        "row_data": {"pdf_url": pdf_url, "text_layer_url": text_layer_url},
        "global_key": str(uuid.uuid4()),
    }


@pytest.fixture()
def conversation_entity_data_row(client, rand_gen):
    return {
        "row_data": "https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json",
        "global_key": f"https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json-{rand_gen(str)}",
    }


@pytest.fixture
def project(client, rand_gen):
    project = client.create_project(
        name=rand_gen(str),
        media_type=MediaType.Image,
    )
    yield project
    project.delete()


@pytest.fixture
def consensus_project(client, rand_gen):
    project = client.create_project(
        name=rand_gen(str),
        quality_modes={QualityMode.Consensus},
        media_type=MediaType.Image,
    )
    yield project
    project.delete()


@pytest.fixture
def model_config(client, rand_gen, valid_model_id):
    model_config = client.create_model_config(
        name=rand_gen(str),
        model_id=valid_model_id,
        inference_params={"param": "value"},
    )
    yield model_config
    client.delete_model_config(model_config.uid)


@pytest.fixture
def consensus_project_with_batch(
    consensus_project, initial_dataset, rand_gen, image_url
):
    project = consensus_project
    dataset = initial_dataset

    data_rows = []
    for _ in range(3):
        data_rows.append(
            {DataRow.row_data: image_url, DataRow.global_key: str(uuid.uuid4())}
        )
    task = dataset.create_data_rows(data_rows)
    task.wait_till_done()
    assert task.status == "COMPLETE"

    data_rows = list(dataset.data_rows())
    assert len(data_rows) == 3
    batch = project.create_batch(
        rand_gen(str),
        data_rows,  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )

    yield [project, batch, data_rows]
    batch.delete()


@pytest.fixture
def dataset(client, rand_gen):
    dataset = client.create_dataset(name=rand_gen(str))
    yield dataset
    dataset.delete()


@pytest.fixture(scope="function")
def unique_dataset(client, rand_gen):
    dataset = client.create_dataset(name=rand_gen(str))
    yield dataset
    dataset.delete()


@pytest.fixture
def small_dataset(dataset: Dataset):
    task = dataset.create_data_rows(
        [
            {"row_data": SMALL_DATASET_URL, "external_id": "my-image"},
        ]
        * 2
    )
    task.wait_till_done()

    yield dataset


@pytest.fixture
def data_row(dataset, image_url, rand_gen):
    global_key = f"global-key-{rand_gen(str)}"
    task = dataset.create_data_rows(
        [
            {
                "row_data": image_url,
                "external_id": "my-image",
                "global_key": global_key,
            },
        ]
    )
    task.wait_till_done()
    dr = dataset.data_rows().get_one()
    yield dr
    dr.delete()


@pytest.fixture
def data_row_and_global_key(dataset, image_url, rand_gen):
    global_key = f"global-key-{rand_gen(str)}"
    task = dataset.create_data_rows(
        [
            {
                "row_data": image_url,
                "external_id": "my-image",
                "global_key": global_key,
            },
        ]
    )
    task.wait_till_done()
    dr = dataset.data_rows().get_one()
    yield dr, global_key
    dr.delete()


# can be used with
# @pytest.mark.parametrize('data_rows', [<count of data rows>], indirect=True)
# if omitted, count defaults to 1
@pytest.fixture
def data_rows(
    dataset, image_url, request, wait_for_data_row_processing, client
):
    count = 1
    if hasattr(request, "param"):
        count = request.param

    datarows = [
        dict(row_data=image_url, global_key=f"global-key-{uuid.uuid4()}")
        for _ in range(count)
    ]

    task = dataset.create_data_rows(datarows)
    task.wait_till_done()
    datarows = dataset.data_rows().get_many(count)
    for dr in dataset.data_rows():
        wait_for_data_row_processing(client, dr)

    yield datarows

    for datarow in datarows:
        datarow.delete()


@pytest.fixture
def iframe_url(environ) -> str:
    if environ in [Environ.PROD, Environ.LOCAL]:
        return "https://editor.labelbox.com"
    elif environ == Environ.STAGING:
        return "https://editor.lb-stage.xyz"


@pytest.fixture
def sample_image() -> str:
    path_to_video = "tests/integration/media/sample_image.jpg"
    return path_to_video


@pytest.fixture
def sample_video() -> str:
    path_to_video = "tests/integration/media/cat.mp4"
    return path_to_video


@pytest.fixture
def sample_bulk_conversation() -> list:
    path_to_conversation = "tests/integration/media/bulk_conversation.json"
    with open(path_to_conversation) as json_file:
        conversations = json.load(json_file)
    return conversations


@pytest.fixture
def organization(client):
    # Must have at least one seat open in your org to run these tests
    org = client.get_organization()

    yield org


@pytest.fixture
def configured_project_with_label(
    client,
    rand_gen,
    dataset,
    data_row,
    wait_for_label_processing,
    teardown_helpers,
):
    """Project with a connected dataset, having one datarow

    Project contains an ontology with 1 bbox tool
    Additionally includes a create_label method for any needed extra labels
    One label is already created and yielded when using fixture
    """
    project = client.create_project(
        name=rand_gen(str),
        media_type=MediaType.Image,
    )
    project._wait_until_data_rows_are_processed(
        data_row_ids=[data_row.uid],
        wait_processing_max_seconds=DATA_ROW_PROCESSING_WAIT_TIMEOUT_SECONDS,
        sleep_interval=DATA_ROW_PROCESSING_WAIT_SLEEP_INTERNAL_SECONDS,
    )

    project.create_batch(
        rand_gen(str),
        [data_row.uid],  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )
    ontology = _setup_ontology(project, client)
    label = _create_label(
        project, data_row, ontology, wait_for_label_processing
    )
    yield [project, dataset, data_row, label]

    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)


def _create_label(project, data_row, ontology, wait_for_label_processing):
    predictions = [
        {
            "uuid": str(uuid.uuid4()),
            "schemaId": ontology.tools[0].feature_schema_id,
            "dataRow": {"id": data_row.uid},
            "bbox": {"top": 20, "left": 20, "height": 50, "width": 50},
        }
    ]

    def create_label():
        """Ad-hoc function to create a LabelImport
        Creates a LabelImport task which will create a label
        """
        upload_task = LabelImport.create_from_objects(
            project.client,
            project.uid,
            f"label-import-{uuid.uuid4()}",
            predictions,
        )
        upload_task.wait_until_done(sleep_time_seconds=5)
        assert (
            upload_task.state == AnnotationImportState.FINISHED
        ), "Label Import did not finish"
        assert (
            len(upload_task.errors) == 0
        ), f"Label Import {upload_task.name} failed with errors {upload_task.errors}"

    project.create_label = create_label
    project.create_label()
    label = wait_for_label_processing(project)[0]
    return label


def _setup_ontology(project: Project, client: Client):
    ontology_builder = OntologyBuilder(
        tools=[
            Tool(tool=Tool.Type.BBOX, name="test-bbox-class"),
        ]
    )
    ontology = client.create_ontology(
        name="ontology with features",
        media_type=MediaType.Image,
        normalized=ontology_builder.asdict(),
    )
    project.connect_ontology(ontology)

    return OntologyBuilder.from_project(project)


@pytest.fixture
def big_dataset(dataset: Dataset):
    task = dataset.create_data_rows(
        [
            {"row_data": IMAGE_URL, "external_id": EXTERNAL_ID},
        ]
        * 3
    )
    task.wait_till_done()

    yield dataset


@pytest.fixture
def configured_batch_project_with_label(
    client,
    dataset,
    data_row,
    wait_for_label_processing,
    rand_gen,
    teardown_helpers,
):
    """Project with a batch having one datarow
    Project contains an ontology with 1 bbox tool
    Additionally includes a create_label method for any needed extra labels
    One label is already created and yielded when using fixture
    """
    project = client.create_project(
        name=rand_gen(str),
        media_type=MediaType.Image,
    )
    data_rows = [dr.uid for dr in list(dataset.data_rows())]
    project._wait_until_data_rows_are_processed(
        data_row_ids=data_rows, sleep_interval=3
    )
    project.create_batch("test-batch", data_rows)
    project.data_row_ids = data_rows

    ontology = _setup_ontology(project, client)
    label = _create_label(
        project, data_row, ontology, wait_for_label_processing
    )

    yield [project, dataset, data_row, label]

    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)


@pytest.fixture
def configured_batch_project_with_multiple_datarows(
    client,
    dataset,
    data_rows,
    wait_for_label_processing,
    rand_gen,
    teardown_helpers,
):
    """Project with a batch having multiple datarows
    Project contains an ontology with 1 bbox tool
    Additionally includes a create_label method for any needed extra labels
    """
    project = client.create_project(
        name=rand_gen(str),
        media_type=MediaType.Image,
    )
    global_keys = [dr.global_key for dr in data_rows]

    batch_name = f"batch {uuid.uuid4()}"
    project.create_batch(batch_name, global_keys=global_keys)

    ontology = _setup_ontology(project, client)
    for datarow in data_rows:
        _create_label(project, datarow, ontology, wait_for_label_processing)

    yield [project, dataset, data_rows]

    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)


# NOTE this is nice heuristics, also there is this logic _wait_until_data_rows_are_processed in Project
#    in case we still have flakiness in the future, we can use it
@pytest.fixture
def wait_for_data_row_processing():
    """
    Do not use. Only for testing.

    Returns DataRow after waiting for it to finish processing media_attributes.
    Some tests, specifically ones that rely on label export, rely on
    DataRow be fully processed with media_attributes
    """

    def func(client, data_row, custom_check=None):
        """
        added check_updated_at because when a data_row is updated from say
        an image to pdf, it already has media_attributes and the loop does
        not wait for processing to a pdf
        """
        data_row_id = data_row.uid
        timeout_seconds = 60
        while True:
            data_row = client.get_data_row(data_row_id)
            passed_custom_check = not custom_check or custom_check(data_row)
            if data_row.media_attributes and passed_custom_check:
                return data_row
            timeout_seconds -= 2
            if timeout_seconds <= 0:
                raise TimeoutError(
                    f"Timed out waiting for DataRow '{data_row_id}' to finish processing media_attributes"
                )
            time.sleep(2)

    return func


@pytest.fixture
def wait_for_label_processing():
    """
    Do not use. Only for testing.

    Returns project's labels as a list after waiting for them to finish processing.
    If `project.labels()` is called before label is fully processed,
    it may return an empty set
    """

    def func(project):
        timeout_seconds = 10
        while True:
            labels = list(project.labels())
            if len(labels) > 0:
                return labels
            timeout_seconds -= 2
            if timeout_seconds <= 0:
                raise TimeoutError(
                    f"Timed out waiting for label for project '{project.uid}' to finish processing"
                )
            time.sleep(2)

    return func


@pytest.fixture
def initial_dataset(client, rand_gen):
    dataset = client.create_dataset(name=rand_gen(str))
    yield dataset

    dataset.delete()


@pytest.fixture
def video_data(client, rand_gen, video_data_row, wait_for_data_row_processing):
    dataset = client.create_dataset(name=rand_gen(str))
    data_row_ids = []
    data_row = dataset.create_data_row(video_data_row)
    data_row = wait_for_data_row_processing(client, data_row)
    data_row_ids.append(data_row.uid)
    yield dataset, data_row_ids
    dataset.delete()


def create_video_data_row(rand_gen):
    return {
        "row_data": "https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4",
        "global_key": f"https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4-{rand_gen(str)}",
        "media_type": "VIDEO",
    }


@pytest.fixture
def video_data_100_rows(client, rand_gen, wait_for_data_row_processing):
    dataset = client.create_dataset(name=rand_gen(str))
    data_row_ids = []
    for _ in range(100):
        data_row = dataset.create_data_row(create_video_data_row(rand_gen))
        data_row = wait_for_data_row_processing(client, data_row)
        data_row_ids.append(data_row.uid)
    yield dataset, data_row_ids
    dataset.delete()


@pytest.fixture()
def video_data_row(rand_gen):
    return create_video_data_row(rand_gen)


class ExportV2Helpers:
    @classmethod
    def run_project_export_v2_task(
        cls, project, num_retries=5, task_name=None, filters={}, params={}
    ):
        task = None
        params = (
            params
            if params
            else {
                "project_details": True,
                "performance_details": False,
                "data_row_details": True,
                "label_details": True,
            }
        )
        while num_retries > 0:
            task = project.export_v2(
                task_name=task_name, filters=filters, params=params
            )
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break
        return task.result

    @classmethod
    def run_dataset_export_v2_task(
        cls, dataset, num_retries=5, task_name=None, filters={}, params={}
    ):
        task = None
        params = (
            params
            if params
            else {"performance_details": False, "label_details": True}
        )
        while num_retries > 0:
            task = dataset.export_v2(
                task_name=task_name, filters=filters, params=params
            )
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break

        return task.result

    @classmethod
    def run_catalog_export_v2_task(
        cls, client, num_retries=5, task_name=None, filters={}, params={}
    ):
        task = None
        params = (
            params
            if params
            else {"performance_details": False, "label_details": True}
        )
        catalog = client.get_catalog()
        while num_retries > 0:
            task = catalog.export_v2(
                task_name=task_name, filters=filters, params=params
            )
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break

        return task.result


@pytest.fixture
def export_v2_test_helpers() -> Type[ExportV2Helpers]:
    return ExportV2Helpers()


@pytest.fixture
def big_dataset_data_row_ids(big_dataset: Dataset):
    export_task = big_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    yield [dr.json["data_row"]["id"] for dr in stream]


@pytest.fixture(scope="function")
def dataset_with_invalid_data_rows(
    unique_dataset: Dataset, upload_invalid_data_rows_for_dataset
):
    upload_invalid_data_rows_for_dataset(unique_dataset)

    yield unique_dataset


@pytest.fixture
def upload_invalid_data_rows_for_dataset():
    def _upload_invalid_data_rows_for_dataset(dataset: Dataset):
        task = dataset.create_data_rows(
            [
                {
                    "row_data": "gs://invalid-bucket/example.png",  # forbidden
                    "external_id": "image-without-access.jpg",
                },
            ]
            * 2
        )
        task.wait_till_done()

    return _upload_invalid_data_rows_for_dataset


@pytest.fixture
def configured_project(
    project_with_one_feature_ontology, initial_dataset, rand_gen, image_url
):
    dataset = initial_dataset
    data_row_id = dataset.create_data_row(row_data=image_url).uid
    project = project_with_one_feature_ontology

    batch = project.create_batch(
        rand_gen(str),
        [data_row_id],  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = [data_row_id]

    yield project

    batch.delete()


@pytest.fixture
def project_with_one_feature_ontology(project, client: Client):
    tools = [
        Tool(tool=Tool.Type.BBOX, name="test-bbox-class").asdict(),
    ]
    empty_ontology = {"tools": tools, "classifications": []}
    ontology = client.create_ontology(
        "empty ontology", empty_ontology, MediaType.Image
    )
    project.connect_ontology(ontology)
    yield project


@pytest.fixture
def configured_project_with_complex_ontology(
    client: Client, initial_dataset, rand_gen, image_url, teardown_helpers
):
    project = client.create_project(
        name=rand_gen(str),
        media_type=MediaType.Image,
    )
    dataset = initial_dataset
    data_row = dataset.create_data_row(row_data=image_url)
    data_row_ids = [data_row.uid]

    project.create_batch(
        rand_gen(str),
        data_row_ids,  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = data_row_ids

    ontology = OntologyBuilder()
    tools = [
        Tool(tool=Tool.Type.BBOX, name="test-bbox-class"),
        Tool(tool=Tool.Type.LINE, name="test-line-class"),
        Tool(tool=Tool.Type.POINT, name="test-point-class"),
        Tool(tool=Tool.Type.POLYGON, name="test-polygon-class"),
    ]

    options = [
        Option(value="first option answer"),
        Option(value="second option answer"),
        Option(value="third option answer"),
    ]

    classifications = [
        Classification(
            class_type=Classification.Type.TEXT, name="test-text-class"
        ),
        Classification(
            class_type=Classification.Type.RADIO,
            name="test-radio-class",
            options=options,
        ),
        Classification(
            class_type=Classification.Type.CHECKLIST,
            name="test-checklist-class",
            options=options,
        ),
    ]

    for t in tools:
        for c in classifications:
            t.add_classification(c)
        ontology.add_tool(t)
    for c in classifications:
        ontology.add_classification(c)

    ontology = client.create_ontology(
        "complex image ontology", ontology.asdict(), MediaType.Image
    )

    project.connect_ontology(ontology)

    yield [project, data_row]
    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)


@pytest.fixture
def embedding(client: Client, environ):
    uuid_str = uuid.uuid4().hex
    time.sleep(randint(1, 5))
    embedding = client.create_embedding(f"sdk-int-{uuid_str}", 8)
    yield embedding

    embedding.delete()


@pytest.fixture
def valid_model_id():
    return "2c903542-d1da-48fd-9db1-8c62571bd3d2"


@pytest.fixture
def requested_labeling_service(
    rand_gen, client, chat_evaluation_ontology, model_config, teardown_helpers
):
    project_name = f"test-model-evaluation-project-{rand_gen(str)}"
    dataset_name = f"test-model-evaluation-dataset-{rand_gen(str)}"
    project = client.create_model_evaluation_project(
        name=project_name, dataset_name=dataset_name, data_row_count=1
    )
    project.connect_ontology(chat_evaluation_ontology)

    project.upsert_instructions("tests/integration/media/sample_pdf.pdf")

    labeling_service = project.get_labeling_service()
    project.add_model_config(model_config.uid)
    project.set_project_model_setup_complete()

    labeling_service.request()

    yield project, project.get_labeling_service()

    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)


class TearDownHelpers:
    @staticmethod
    def teardown_project_labels_ontology_feature_schemas(project: Project):
        """
        Call this function to release project, labels, ontology and feature schemas in fixture teardown

        NOTE: exception handling is not required as this is a fixture teardown
        """
        ontology = project.ontology()
        ontology_id = ontology.uid
        client = project.client
        classification_feature_schema_ids = [
            feature["featureSchemaId"]
            for feature in ontology.normalized["classifications"]
        ]
        tool_feature_schema_ids = [
            feature["featureSchemaId"]
            for feature in ontology.normalized["tools"]
        ]

        feature_schema_ids = (
            classification_feature_schema_ids + tool_feature_schema_ids
        )
        labels = list(project.labels())
        for label in labels:
            label.delete()

        project.delete()
        client.delete_unused_ontology(ontology_id)
        for feature_schema_id in feature_schema_ids:
            try:
                project.client.delete_unused_feature_schema(feature_schema_id)
            except LabelboxError as e:
                print(
                    f"Failed to delete feature schema {feature_schema_id}: {e}"
                )

    @staticmethod
    def teardown_ontology_feature_schemas(ontology: Ontology):
        """
        Call this function to release project, labels, ontology and feature schemas in fixture teardown

        NOTE: exception handling is not required as this is a fixture teardown
        """
        ontology_id = ontology.uid
        client = ontology.client
        classification_feature_schema_ids = [
            feature["featureSchemaId"]
            for feature in ontology.normalized["classifications"]
        ] + [
            option["featureSchemaId"]
            for feature in ontology.normalized["classifications"]
            for option in feature.get("options", [])
        ]

        tool_feature_schema_ids = (
            [
                feature["featureSchemaId"]
                for feature in ontology.normalized["tools"]
            ]
            + [
                classification["featureSchemaId"]
                for tool in ontology.normalized["tools"]
                for classification in tool.get("classifications", [])
            ]
            + [
                option["featureSchemaId"]
                for tool in ontology.normalized["tools"]
                for classification in tool.get("classifications", [])
                for option in classification.get("options", [])
            ]
        )

        feature_schema_ids = (
            classification_feature_schema_ids + tool_feature_schema_ids
        )

        client.delete_unused_ontology(ontology_id)
        for feature_schema_id in feature_schema_ids:
            try:
                project.client.delete_unused_feature_schema(feature_schema_id)
            except LabelboxError as e:
                print(
                    f"Failed to delete feature schema {feature_schema_id}: {e}"
                )


class ModuleTearDownHelpers(TearDownHelpers): ...


class LabelHelpers:
    def wait_for_labels(self, project, number_of_labels=1):
        timeout_seconds = 10
        while True:
            labels = list(project.labels())
            if len(labels) >= number_of_labels:
                return labels
            timeout_seconds -= 2
            if timeout_seconds <= 0:
                raise TimeoutError(
                    f"Timed out waiting for label for project '{project.uid}' to finish processing"
                )
            time.sleep(2)


@pytest.fixture
def teardown_helpers():
    return TearDownHelpers()


@pytest.fixture(scope="module")
def module_teardown_helpers():
    return TearDownHelpers()


@pytest.fixture
def label_helpers():
    return LabelHelpers()



from typing import List, Union

from labelbox.schema.id_type import IdType


class Identifiables:
    def __init__(self, iterable, id_type: str):
        """
        Args:
            iterable: Iterable of ids (unique or global keys)
            id_type: The type of id used to identify a data row.
        """
        self._iterable = iterable
        self._id_type = id_type

    @property
    def id_type(self):
        return self._id_type

    def __iter__(self):
        return iter(self._iterable)

    def __getitem__(self, index):
        if isinstance(index, slice):
            ids = self._iterable[index]
            return self.__class__(ids)  # type: ignore
        return self._iterable[index]

    def __len__(self):
        return len(self._iterable)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._iterable})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Identifiables):
            return False
        return (
            self._iterable == other._iterable
            and self._id_type == other._id_type
        )


class UniqueIds(Identifiables):
    """
    Represents a collection of unique, internally generated ids.
    """

    def __init__(self, iterable: List[str]):
        super().__init__(iterable, IdType.DataRowId)


class GlobalKeys(Identifiables):
    """
    Represents a collection of user generated ids.
    """

    def __init__(self, iterable: List[str]):
        super().__init__(iterable, IdType.GlobalKey)


DataRowIds = UniqueIds

DataRowIdentifiers = Union[UniqueIds, GlobalKeys]



from typing import Optional, Tuple, Union

import geojson
import numpy as np
import cv2
from shapely.geometry import Point as SPoint

from .geometry import Geometry


class Point(Geometry):
    """Point geometry

    >>> Point(x=0, y=0)

    Args:
        x (float)
        y (float)

    """

    x: float
    y: float

    @property
    def geometry(self) -> geojson.Point:
        return geojson.Point((self.x, self.y))

    @classmethod
    def from_shapely(cls, shapely_obj: SPoint) -> "Point":
        """Transforms a shapely object."""
        if not isinstance(shapely_obj, SPoint):
            raise TypeError(
                f"Expected Shapely Point. Got {shapely_obj.geom_type}"
            )

        obj_coords = shapely_obj.__geo_interface__["coordinates"]
        return Point(x=obj_coords[0], y=obj_coords[1])

    def draw(
        self,
        height: Optional[int] = None,
        width: Optional[int] = None,
        canvas: Optional[np.ndarray] = None,
        color: Union[int, Tuple[int, int, int]] = (255, 255, 255),
        thickness: int = 10,
    ) -> np.ndarray:
        """
        Draw the point onto a 3d mask
        Args:
            height (int): height of the mask
            width (int): width of the mask
            thickness (int): pixel radius of the point
            color (int): color for the point.
                  RGB values by default but if a 2D canvas is provided this can set this to an int.
            canvas (np.ndarray): Canvas to draw the point on
        Returns:
            numpy array representing the mask with the point drawn on it.
        """
        canvas = self.get_or_create_canvas(height, width, canvas)
        return cv2.circle(
            canvas,
            (int(self.x), int(self.y)),
            radius=thickness,
            color=color,
            thickness=-1,
        )



from enum import Enum
from typing import Dict, Optional

from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field


class AttachmentType(str, Enum):
    VIDEO = "VIDEO"
    IMAGE = "IMAGE"
    IMAGE_OVERLAY = "IMAGE_OVERLAY"
    HTML = "HTML"
    RAW_TEXT = "RAW_TEXT"
    TEXT_URL = "TEXT_URL"
    PDF_URL = "PDF_URL"
    CAMERA_IMAGE = "CAMERA_IMAGE"  # Used by experimental point-cloud editor


class AssetAttachment(DbObject):
    """Asset attachment provides extra context about an asset while labeling.

    Attributes:
        attachment_type (str): IMAGE, VIDEO, IMAGE_OVERLAY, HTML, RAW_TEXT, TEXT_URL, or PDF_URL.
        attachment_value (str): URL to an external file or a string of text
        attachment_name (str): The name of the attachment
    """

    for topic in AttachmentType:
        vars()[topic.name] = topic.value

    attachment_type = Field.String("attachment_type", "type")
    attachment_value = Field.String("attachment_value", "value")
    attachment_name = Field.String("attachment_name", "name")

    @classmethod
    def validate_attachment_json(cls, attachment_json: Dict[str, str]) -> None:
        for required_key in ["type", "value"]:
            if required_key not in attachment_json:
                raise ValueError(
                    f"Must provide a `{required_key}` key for each attachment. Found {attachment_json}."
                )
        cls.validate_attachment_value(attachment_json["value"])
        cls.validate_attachment_type(attachment_json["type"])

    @classmethod
    def validate_attachment_value(cls, attachment_value: str) -> None:
        if not isinstance(attachment_value, str) or attachment_value == "":
            raise ValueError(
                f"Attachment value must be a non-empty string, got: '{attachment_value}'"
            )

    @classmethod
    def validate_attachment_type(cls, attachment_type: str) -> None:
        valid_types = set(AttachmentType.__members__)
        if attachment_type not in valid_types:
            raise ValueError(
                f"attachment_type must be one of {valid_types}. Found {attachment_type}"
            )

    def delete(self) -> None:
        """Deletes an attachment on the data row."""
        query_str = """mutation deleteDataRowAttachmentPyApi($attachment_id: ID!) {
            deleteDataRowAttachment(where: {id: $attachment_id}) {
                    id}
            }"""
        self.client.execute(query_str, {"attachment_id": self.uid})

    def update(
        self,
        name: Optional[str] = None,
        type: Optional[str] = None,
        value: Optional[str] = None,
    ):
        """Updates an attachment on the data row."""
        if not name and not type and value is None:
            raise ValueError(
                "At least one of the following must be provided: name, type, value"
            )

        query_params = {"attachment_id": self.uid}
        if type:
            self.validate_attachment_type(type)
            query_params["type"] = type
        if value is not None:
            self.validate_attachment_value(value)
            query_params["value"] = value
        if name:
            query_params["name"] = name

        query_str = """mutation updateDataRowAttachmentPyApi($attachment_id: ID!, $name: String, $type: AttachmentType, $value: String) {
            updateDataRowAttachment(
              where: {id: $attachment_id}, 
              data: {name: $name, type: $type, value: $value}
            ) { id name type value }
            }"""
        res = (self.client.execute(query_str, query_params))[
            "updateDataRowAttachment"
        ]

        self.attachment_name = res["name"]
        self.attachment_value = res["value"]
        self.attachment_type = res["type"]



import warnings
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from lbox.exceptions import InconsistentOntologyException

from labelbox.schema.tool_building.types import (
    FeatureSchemaId,
    FeatureSchemaAttributes,
    FeatureSchemaAttribute,
)


@dataclass
class Classification:
    """
    A classification to be added to a Project's ontology. The
    classification is dependent on the Classification Type.

    To instantiate, the "class_type" and "name" parameters must
    be passed in.

    The "options" parameter holds a list of Option objects. This is not
    necessary for some Classification types, such as TEXT. To see which
    types require options, look at the "_REQUIRES_OPTIONS" class variable.

    Example(s):
        classification = Classification(
            class_type = Classification.Type.TEXT,
            name = "Classification Example")

        classification_two = Classification(
            class_type = Classification.Type.RADIO,
            name = "Second Example")
        classification_two.add_option(Option(
            value = "Option Example"))

    Attributes:
        class_type: (Classification.Type)
        name: (str)
        instructions: (str)
        required: (bool)
        options: (list)
        ui_mode: (str)
        schema_id: (str)
        feature_schema_id: (str)
        scope: (str)
        attributes: (list)
    """

    class Type(Enum):
        TEXT = "text"
        CHECKLIST = "checklist"
        RADIO = "radio"

    class Scope(Enum):
        GLOBAL = "global"
        INDEX = "index"

    class UIMode(Enum):
        HOTKEY = "hotkey"
        SEARCHABLE = "searchable"

    _REQUIRES_OPTIONS = {Type.CHECKLIST, Type.RADIO}

    class_type: Type
    name: Optional[str] = None
    instructions: Optional[str] = None
    required: bool = False
    options: List["Option"] = field(default_factory=list)
    schema_id: Optional[str] = None
    feature_schema_id: Optional[str] = None
    scope: Optional[Scope] = None
    ui_mode: Optional[UIMode] = (
        None  # How this classification should be answered (e.g. hotkeys / autocomplete, etc)
    )
    attributes: Optional[FeatureSchemaAttributes] = None

    def __post_init__(self):
        if self.name is None:
            msg = (
                "When creating the Classification feature, please use “name” "
                "for the classification schema name, which will be used when "
                "creating annotation payload for Model-Assisted Labeling "
                "Import and Label Import. “instructions” is no longer "
                "supported to specify classification schema name."
            )
            if self.instructions is not None:
                self.name = self.instructions
                warnings.warn(msg)
            else:
                raise ValueError(msg)
        else:
            if self.instructions is None:
                self.instructions = self.name
        if self.attributes is not None:
            warnings.warn(
                "The attributes for Classifications are in beta. The attribute name and signature may change in the future."
            )

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> "Classification":
        return cls(
            class_type=Classification.Type(dictionary["type"]),
            name=dictionary["name"],
            instructions=dictionary["instructions"],
            required=dictionary.get("required", False),
            options=[Option.from_dict(o) for o in dictionary["options"]],
            ui_mode=cls.UIMode(dictionary["uiMode"])
            if "uiMode" in dictionary
            else None,
            schema_id=dictionary.get("schemaNodeId", None),
            feature_schema_id=dictionary.get("featureSchemaId", None),
            scope=cls.Scope(dictionary.get("scope", cls.Scope.GLOBAL)),
            attributes=[
                FeatureSchemaAttribute.from_dict(attr)
                for attr in dictionary.get("attributes", []) or []
            ]
            if dictionary.get("attributes")
            else None,
        )

    def asdict(self, is_subclass: bool = False) -> Dict[str, Any]:
        if self.class_type in self._REQUIRES_OPTIONS and len(self.options) < 1:
            raise InconsistentOntologyException(
                f"Classification '{self.name}' requires options."
            )
        classification = {
            "type": self.class_type.value,
            "instructions": self.instructions,
            "name": self.name,
            "required": self.required,
            "options": [o.asdict() for o in self.options],
            "schemaNodeId": self.schema_id,
            "featureSchemaId": self.feature_schema_id,
            "attributes": [a.asdict() for a in self.attributes]
            if self.attributes is not None
            else None,
        }
        if (
            self.class_type == self.Type.RADIO
            or self.class_type == self.Type.CHECKLIST
        ) and self.ui_mode:
            # added because this key does nothing for text so no point of including
            classification["uiMode"] = self.ui_mode.value
        if is_subclass:
            return classification
        classification["scope"] = (
            self.scope.value
            if self.scope is not None
            else self.Scope.GLOBAL.value
        )
        return classification

    def add_option(self, option: "Option") -> None:
        if option.value in (o.value for o in self.options):
            raise InconsistentOntologyException(
                f"Duplicate option '{option.value}' "
                f"for classification '{self.name}'."
            )
        self.options.append(option)


@dataclass
class Option:
    """
    An option is a possible answer within a Classification object in
    a Project's ontology.

    To instantiate, only the "value" parameter needs to be passed in.

    Example(s):
        option = Option(value = "Option Example")

    Attributes:
        value: (str)
        schema_id: (str)
        feature_schema_id: (str)
        options: (list)
    """

    value: Union[str, int]
    label: Optional[Union[str, int]] = None
    schema_id: Optional[str] = None
    feature_schema_id: Optional[FeatureSchemaId] = None  # type: ignore
    options: Union[
        List["Classification"], List["PromptResponseClassification"]
    ] = field(default_factory=list)

    def __post_init__(self):
        if self.label is None:
            self.label = self.value

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> "Option":
        return cls(
            value=dictionary["value"],
            label=dictionary["label"],
            schema_id=dictionary.get("schemaNodeId", None),
            feature_schema_id=dictionary.get("featureSchemaId", None),
            options=[
                Classification.from_dict(o)
                for o in dictionary.get("options", [])
            ],
        )

    def asdict(self) -> Dict[str, Any]:
        return {
            "schemaNodeId": self.schema_id,
            "featureSchemaId": self.feature_schema_id,
            "label": self.label,
            "value": self.value,
            "options": [o.asdict(is_subclass=True) for o in self.options],
        }

    def add_option(
        self, option: Union["Classification", "PromptResponseClassification"]
    ) -> None:
        if option.name in (o.name for o in self.options):
            raise InconsistentOntologyException(
                f"Duplicate nested classification '{option.name}' "
                f"for option '{self.label}'"
            )
        self.options.append(option)  # type: ignore


@dataclass
class PromptResponseClassification:
    """

    A PromptResponseClassification to be added to a Project's ontology. The
    classification is dependent on the PromptResponseClassification Type.

    To instantiate, the "class_type" and "name" parameters must
    be passed in.

    The "options" parameter holds a list of Response Option objects. This is not
    necessary for some Classification types, such as RESPONSE_TEXT or PROMPT. To see which
    types require options, look at the "_REQUIRES_OPTIONS" class variable.

    Example(s):
    >>>    classification = PromptResponseClassification(
    >>>        class_type = PromptResponseClassification.Type.Prompt,
    >>>        character_min = 1,
    >>>        character_max = 1
    >>>        name = "Prompt Classification Example")

    >>>    classification_two = PromptResponseClassification(
    >>>        class_type = PromptResponseClassification.Type.RESPONSE_RADIO,
    >>>        name = "Second Example")

    >>>    classification_two.add_option(ResponseOption(
    >>>        value = "Option Example"))

    Attributes:
        class_type: (Classification.Type)
        name: (str)
        instructions: (str)
        required: (bool)
        options: (list)
        character_min: (int)
        character_max: (int)
        schema_id: (str)
        feature_schema_id: (str)
    """

    def __post_init__(self):
        if self.name is None:
            msg = (
                "When creating the Classification feature, please use “name” "
                "for the classification schema name, which will be used when "
                "creating annotation payload for Model-Assisted Labeling "
                "Import and Label Import. “instructions” is no longer "
                "supported to specify classification schema name."
            )
            if self.instructions is not None:
                self.name = self.instructions
                warnings.warn(msg)
            else:
                raise ValueError(msg)
        else:
            if self.instructions is None:
                self.instructions = self.name

    class Type(Enum):
        PROMPT = "prompt"
        RESPONSE_TEXT = "response-text"
        RESPONSE_CHECKLIST = "response-checklist"
        RESPONSE_RADIO = "response-radio"

    _REQUIRES_OPTIONS = {Type.RESPONSE_CHECKLIST, Type.RESPONSE_RADIO}

    class_type: Type
    name: Optional[str] = None
    instructions: Optional[str] = None
    required: bool = True
    options: List["ResponseOption"] = field(default_factory=list)
    character_min: Optional[int] = None
    character_max: Optional[int] = None
    schema_id: Optional[str] = None
    feature_schema_id: Optional[str] = None

    @classmethod
    def from_dict(
        cls, dictionary: Dict[str, Any]
    ) -> "PromptResponseClassification":
        return cls(
            class_type=PromptResponseClassification.Type(dictionary["type"]),
            name=dictionary["name"],
            instructions=dictionary["instructions"],
            required=True,  # always required
            options=[
                ResponseOption.from_dict(o) for o in dictionary["options"]
            ],
            character_min=dictionary.get("minCharacters", None),
            character_max=dictionary.get("maxCharacters", None),
            schema_id=dictionary.get("schemaNodeId", None),
            feature_schema_id=dictionary.get("featureSchemaId", None),
        )

    def asdict(self, is_subclass: bool = False) -> Dict[str, Any]:
        if self.class_type in self._REQUIRES_OPTIONS and len(self.options) < 1:
            raise InconsistentOntologyException(
                f"Response Classification '{self.name}' requires options."
            )
        classification: Dict[str, Any] = {
            "type": self.class_type.value,
            "instructions": self.instructions,
            "name": self.name,
            "required": True,  # always required
            "options": [o.asdict() for o in self.options],
            "schemaNodeId": self.schema_id,
            "featureSchemaId": self.feature_schema_id,
        }
        if (
            self.class_type == self.Type.PROMPT
            or self.class_type == self.Type.RESPONSE_TEXT
        ):
            if self.character_min:
                classification["minCharacters"] = self.character_min
            if self.character_max:
                classification["maxCharacters"] = self.character_max
        if is_subclass:
            return classification
        return classification

    def add_option(self, option: "ResponseOption") -> None:
        if option.value in (o.value for o in self.options):
            raise InconsistentOntologyException(
                f"Duplicate option '{option.value}' "
                f"for response classification '{self.name}'."
            )
        self.options.append(option)


@dataclass
class ResponseOption(Option):
    """
    An option is a possible answer within a PromptResponseClassification response object in
    a Project's ontology.

    To instantiate, only the "value" parameter needs to be passed in.

    Example(s):
        option = ResponseOption(value = "Response Option Example")

    Attributes:
        value: (str)
        schema_id: (str)
        feature_schema_id: (str)
        options: (list)
    """

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> "ResponseOption":
        return cls(
            value=dictionary["value"],
            label=dictionary["label"],
            schema_id=dictionary.get("schemaNodeId", None),
            feature_schema_id=dictionary.get("featureSchemaId", None),
            options=[
                PromptResponseClassification.from_dict(o)
                for o in dictionary.get("options", [])
            ],
        )



import json

from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
from labelbox.data.mixins import CustomMetric
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.types import (
    Label,
    ObjectAnnotation,
    Rectangle,
    Point,
    ClassificationAnnotation,
    Radio,
    ClassificationAnswer,
    Text,
    Checklist,
)


def test_nested():
    with open("tests/data/assets/ndjson/nested_import.json", "r") as file:
        data = json.load(file)
    labels = [
        Label(
            data=GenericDataRowData(
                uid="ckrb1sf1i1g7i0ybcdc6oc8ct",
            ),
            annotations=[
                ObjectAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={
                        "uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673",
                    },
                    value=Rectangle(
                        start=Point(x=2275.0, y=1352.0),
                        end=Point(x=2414.0, y=1702.0),
                    ),
                    classifications=[
                        ClassificationAnnotation(
                            feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                            value=Radio(
                                answer=ClassificationAnswer(
                                    custom_metrics=[
                                        CustomMetric(
                                            name="customMetric1", value=0.5
                                        ),
                                        CustomMetric(
                                            name="customMetric2", value=0.3
                                        ),
                                    ],
                                    confidence=0.34,
                                    feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                                ),
                            ),
                        )
                    ],
                ),
                ObjectAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={
                        "uuid": "d009925d-91a3-4f67-abd9-753453f5a584",
                    },
                    value=Rectangle(
                        start=Point(x=2089.0, y=1251.0),
                        end=Point(x=2247.0, y=1679.0),
                    ),
                    classifications=[
                        ClassificationAnnotation(
                            feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                            value=Radio(
                                answer=ClassificationAnswer(
                                    feature_schema_id="ckrb1sfl8099e0y919v260awv",
                                ),
                            ),
                        )
                    ],
                ),
                ObjectAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={
                        "uuid": "5d03213e-4408-456c-9eca-cf0723202961",
                    },
                    value=Rectangle(
                        start=Point(x=2089.0, y=1251.0),
                        end=Point(x=2247.0, y=1679.0),
                    ),
                    classifications=[
                        ClassificationAnnotation(
                            feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                            value=Checklist(
                                answer=[
                                    ClassificationAnswer(
                                        custom_metrics=[
                                            CustomMetric(
                                                name="customMetric1", value=0.5
                                            ),
                                            CustomMetric(
                                                name="customMetric2", value=0.3
                                            ),
                                        ],
                                        confidence=0.894,
                                        feature_schema_id="ckrb1sfl8099e0y919v260awv",
                                    )
                                ],
                            ),
                        )
                    ],
                ),
                ObjectAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={
                        "uuid": "d50812f6-34eb-4f12-b3cb-bbde51a31d83",
                    },
                    value=Rectangle(
                        start=Point(x=2089.0, y=1251.0),
                        end=Point(x=2247.0, y=1679.0),
                    ),
                    classifications=[
                        ClassificationAnnotation(
                            feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                            extra={},
                            value=Text(
                                answer="a string",
                            ),
                        )
                    ],
                ),
            ],
        )
    ]
    res = list(NDJsonConverter.serialize(labels))
    assert res == data


def test_nested_name_only():
    with open(
        "tests/data/assets/ndjson/nested_import_name_only.json", "r"
    ) as file:
        data = json.load(file)
    labels = [
        Label(
            data=GenericDataRowData(
                uid="ckrb1sf1i1g7i0ybcdc6oc8ct",
            ),
            annotations=[
                ObjectAnnotation(
                    name="box a",
                    extra={
                        "uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673",
                    },
                    value=Rectangle(
                        start=Point(x=2275.0, y=1352.0),
                        end=Point(x=2414.0, y=1702.0),
                    ),
                    classifications=[
                        ClassificationAnnotation(
                            name="classification a",
                            value=Radio(
                                answer=ClassificationAnswer(
                                    custom_metrics=[
                                        CustomMetric(
                                            name="customMetric1", value=0.5
                                        ),
                                        CustomMetric(
                                            name="customMetric2", value=0.3
                                        ),
                                    ],
                                    confidence=0.811,
                                    name="first answer",
                                ),
                            ),
                        )
                    ],
                ),
                ObjectAnnotation(
                    name="box b",
                    extra={
                        "uuid": "d009925d-91a3-4f67-abd9-753453f5a584",
                    },
                    value=Rectangle(
                        start=Point(x=2089.0, y=1251.0),
                        end=Point(x=2247.0, y=1679.0),
                    ),
                    classifications=[
                        ClassificationAnnotation(
                            name="classification b",
                            value=Radio(
                                answer=ClassificationAnswer(
                                    custom_metrics=[
                                        CustomMetric(
                                            name="customMetric1", value=0.5
                                        ),
                                        CustomMetric(
                                            name="customMetric2", value=0.3
                                        ),
                                    ],
                                    confidence=0.815,
                                    name="second answer",
                                ),
                            ),
                        )
                    ],
                ),
                ObjectAnnotation(
                    name="box c",
                    extra={
                        "uuid": "8a2b2c43-f0a1-4763-ba96-e322d986ced6",
                    },
                    value=Rectangle(
                        start=Point(x=2089.0, y=1251.0),
                        end=Point(x=2247.0, y=1679.0),
                    ),
                    classifications=[
                        ClassificationAnnotation(
                            name="classification c",
                            value=Checklist(
                                answer=[
                                    ClassificationAnswer(
                                        name="third answer",
                                    )
                                ],
                            ),
                        )
                    ],
                ),
                ObjectAnnotation(
                    name="box c",
                    extra={
                        "uuid": "456dd2c6-9fa0-42f9-9809-acc27b9886a7",
                    },
                    value=Rectangle(
                        start=Point(x=2089.0, y=1251.0),
                        end=Point(x=2247.0, y=1679.0),
                    ),
                    classifications=[
                        ClassificationAnnotation(
                            name="a string",
                            value=Text(
                                answer="a string",
                            ),
                        )
                    ],
                ),
            ],
        )
    ]
    res = list(NDJsonConverter.serialize(labels))
    assert res == data



from typing import Optional, Union, Type

from labelbox.data.annotation_types.data import GenericDataRowData
from labelbox.data.serialization.ndjson.base import DataRow, NDJsonBase
from labelbox.data.annotation_types.metrics.scalar import (
    ScalarMetric,
    ScalarMetricAggregation,
    ScalarMetricValue,
    ScalarMetricConfidenceValue,
)
from labelbox.data.annotation_types.metrics.confusion_matrix import (
    ConfusionMatrixAggregation,
    ConfusionMatrixMetric,
    ConfusionMatrixMetricValue,
    ConfusionMatrixMetricConfidenceValue,
)
from pydantic import ConfigDict, model_serializer


class BaseNDMetric(NDJsonBase):
    metric_value: float
    feature_name: Optional[str] = None
    subclass_name: Optional[str] = None
    model_config = ConfigDict(use_enum_values=True)

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        res = handler(self)
        for field in ["featureName", "subclassName"]:
            if field in res and res[field] is None:
                res.pop(field)
        return res


class NDConfusionMatrixMetric(BaseNDMetric):
    metric_value: Union[
        ConfusionMatrixMetricValue, ConfusionMatrixMetricConfidenceValue
    ]
    metric_name: str
    aggregation: ConfusionMatrixAggregation

    def to_common(self) -> ConfusionMatrixMetric:
        return ConfusionMatrixMetric(
            value=self.metric_value,
            metric_name=self.metric_name,
            feature_name=self.feature_name,
            subclass_name=self.subclass_name,
            aggregation=self.aggregation,
            extra={"uuid": self.uuid},
        )

    @classmethod
    def from_common(
        cls, metric: ConfusionMatrixMetric, data: GenericDataRowData
    ) -> "NDConfusionMatrixMetric":
        return cls(
            uuid=metric.extra.get("uuid"),
            metric_value=metric.value,
            metric_name=metric.metric_name,
            feature_name=metric.feature_name,
            subclass_name=metric.subclass_name,
            aggregation=metric.aggregation,
            data_row=DataRow(id=data.uid, global_key=data.global_key),
        )


class NDScalarMetric(BaseNDMetric):
    metric_value: Union[ScalarMetricValue, ScalarMetricConfidenceValue]
    metric_name: Optional[str] = None
    aggregation: Optional[ScalarMetricAggregation] = (
        ScalarMetricAggregation.ARITHMETIC_MEAN
    )

    def to_common(self) -> ScalarMetric:
        return ScalarMetric(
            value=self.metric_value,
            metric_name=self.metric_name,
            feature_name=self.feature_name,
            subclass_name=self.subclass_name,
            aggregation=self.aggregation,
            extra={"uuid": self.uuid},
        )

    @classmethod
    def from_common(
        cls, metric: ScalarMetric, data: GenericDataRowData
    ) -> "NDScalarMetric":
        return cls(
            uuid=metric.extra.get("uuid"),
            metric_value=metric.value,
            metric_name=metric.metric_name,
            feature_name=metric.feature_name,
            subclass_name=metric.subclass_name,
            aggregation=metric.aggregation.value,
            data_row=DataRow(id=data.uid, global_key=data.global_key),
        )


class NDMetricAnnotation:
    @classmethod
    def to_common(
        cls, annotation: Union[NDScalarMetric, NDConfusionMatrixMetric]
    ) -> Union[ScalarMetric, ConfusionMatrixMetric]:
        return annotation.to_common()

    @classmethod
    def from_common(
        cls,
        annotation: Union[ScalarMetric, ConfusionMatrixMetric],
        data: GenericDataRowData,
    ) -> Union[NDScalarMetric, NDConfusionMatrixMetric]:
        obj = cls.lookup_object(annotation)
        return obj.from_common(annotation, data)

    @staticmethod
    def lookup_object(
        annotation: Union[ScalarMetric, ConfusionMatrixMetric],
    ) -> Union[Type[NDScalarMetric], Type[NDConfusionMatrixMetric]]:
        result = {
            ScalarMetric: NDScalarMetric,
            ConfusionMatrixMetric: NDConfusionMatrixMetric,
        }.get(type(annotation))
        if result is None:
            raise TypeError(
                f"Unable to convert object to MAL format. `{type(annotation)}`"
            )
        return result



import cv2
import pytest

from labelbox.data.annotation_types import Point, Rectangle
from pydantic import ValidationError


def test_rectangle():
    with pytest.raises(ValidationError):
        rectangle = Rectangle()

    rectangle = Rectangle(start=Point(x=0, y=1), end=Point(x=10, y=10))
    points = [[[0.0, 1.0], [0.0, 10.0], [10.0, 10.0], [10.0, 1.0], [0.0, 1.0]]]
    expected = {"coordinates": points, "type": "Polygon"}
    assert rectangle.geometry == expected
    expected["coordinates"] = tuple([tuple([tuple(x) for x in points[0]])])
    assert rectangle.shapely.__geo_interface__ == expected

    raster = rectangle.draw(height=32, width=32)
    assert (cv2.imread("tests/data/assets/rectangle.png") == raster).all()

    xyhw = Rectangle.from_xyhw(0.0, 0, 10, 10)
    assert xyhw.start == Point(x=0, y=0.0)
    assert xyhw.end == Point(x=10, y=10.0)



from labelbox.schema.ontology_kind import (
    OntologyKind,
    EditorTaskType,
    EditorTaskTypeMapper,
)
from labelbox.schema.media_type import MediaType


def test_ontology_kind_conversions_from_editor_task_type():
    ontology_kind = OntologyKind.ModelEvaluation
    media_type = MediaType.Conversational
    editor_task_type = EditorTaskTypeMapper.to_editor_task_type(
        ontology_kind, media_type
    )
    assert editor_task_type == EditorTaskType.ModelChatEvaluation

    ontology_kind = OntologyKind.Missing
    media_type = MediaType.Image
    editor_task_type = EditorTaskTypeMapper.to_editor_task_type(
        ontology_kind, media_type
    )
    assert editor_task_type == EditorTaskType.Missing

    ontology_kind = OntologyKind.ModelEvaluation
    media_type = MediaType.Video
    editor_task_type = EditorTaskTypeMapper.to_editor_task_type(
        ontology_kind, media_type
    )
    assert editor_task_type == EditorTaskType.Missing



from abc import ABC
from typing import Dict, Optional, Any, Union

from pydantic import (
    confloat,
    BaseModel,
    model_serializer,
    field_validator,
)

ConfidenceValue = confloat(ge=0, le=1)

MIN_CONFIDENCE_SCORES = 2
MAX_CONFIDENCE_SCORES = 15


class BaseMetric(BaseModel, ABC):
    value: Union[Any, Dict[float, Any]]
    feature_name: Optional[str] = None
    subclass_name: Optional[str] = None
    extra: Dict[str, Any] = {}

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        res = handler(self)
        return {k: v for k, v in res.items() if v is not None}

    @field_validator("value")
    def validate_value(cls, value):
        if isinstance(value, Dict):
            if not (
                MIN_CONFIDENCE_SCORES <= len(value) <= MAX_CONFIDENCE_SCORES
            ):
                raise ValueError(
                    f"Number of confidence scores must be greater than\n \
                        or equal to {MIN_CONFIDENCE_SCORES} and less than\n \
                        or equal to {MAX_CONFIDENCE_SCORES}. Found {len(value)}"
                )
        return value



from enum import Enum, auto

""" Classes for defining the client-side comparison operations used
for filtering data in fetches. Intended for use by library internals
and not by the end user.
"""


class LogicalExpressionComponent:
    """Implements bitwise logical operator methods (&, | and ~) so they
    return a LogicalExpression object containing this
    LogicalExpressionComponent.
    """

    def __and__(self, other):
        if not isinstance(other, (LogicalExpression, Comparison)):
            return NotImplemented
        return LogicalExpression.Op.AND(self, other)

    def __or__(self, other):
        if not isinstance(other, (LogicalExpression, Comparison)):
            return NotImplemented
        return LogicalExpression.Op.OR(self, other)

    def __invert__(self):
        return LogicalExpression.Op.NOT(self)


class LogicalExpression(LogicalExpressionComponent):
    """A unary (NOT) or binary (AND, OR) logical expression between
    Comparison or LogicalExpression objects."""

    class Op(Enum):
        """Type of logical operation."""

        AND = auto()
        OR = auto()
        NOT = auto()

        def __call__(self, first, second=None):
            """Forwards to LogicalExpression constructor, passing `self`
            as the `op` argument."""
            return LogicalExpression(self, first, second)

    def __init__(self, op, first, second=None):
        """LogicalExpression constructor.

        Args:
            op (LogicalExpression.Op): The type of logical operation.
            first (LogicalExpression or Comparison): First operand.
            second (LogicalExpression or Comparison): Second operand.
        """
        self.op = op
        self.first = first
        self.second = second

    def __eq__(self, other):
        return self.op == other.op and (
            (self.first == other.first and self.second == other.second)
            or (self.first == other.second and self.second == other.first)
        )

    def __hash__(self):
        return (
            hash(self.op) + 2833 * hash(self.first) + 2837 * hash(self.second)
        )

    def __repr__(self):
        return "%r %s %r" % (self.first, self.op.name, self.second)

    def __str__(self):
        return "%s %s %s" % (self.first, self.op.name, self.second)


class Comparison(LogicalExpressionComponent):
    """A comparison between a database value (represented by a
    `labelbox.schema.Field` object) and a constant value."""

    class Op(Enum):
        """Type of the comparison operation."""

        EQ = auto()
        NE = auto()
        LT = auto()
        GT = auto()
        LE = auto()
        GE = auto()

        def __call__(self, *args):
            """Forwards to Comparison constructor, passing `self`
            as the `op` argument."""
            return Comparison(self, *args)

    def __init__(self, op, field, value):
        """Comparison constructor.

        Args:
            op (Comparison.Op): The type of comparison.
            field (labelbox.schema.Field): Field being compared.
            value (any): Value to which the DB field is compared.
        """
        self.op = op
        self.field = field
        self.value = value

    def __eq__(self, other):
        return (
            self.op == other.op
            and self.field == other.field
            and self.value == other.value
        )

    def __hash__(self):
        return hash(self.op) + 2861 * hash(self.field) + 2927 * hash(self.value)

    def __repr__(self):
        return "%r %s %r" % (self.field, self.op.name, self.value)

    def __str__(self):
        return "%s %s %s" % (self.field, self.op.name, self.value)



import pytest

from labelbox import Project


@pytest.mark.xfail(reason="Sorting not supported on top-level fetches")
def test_top_level_sorting(client):
    client.get_projects(order_by=Project.name.asc)



import pytest

from labelbox.schema.tool_building.classification import Classification
from labelbox.schema.tool_building.prompt_issue_tool import PromptIssueTool


def test_as_dict():
    tool = PromptIssueTool(name="Prompt Issue Tool")

    # Get the dictionary representation
    tool_dict = tool.asdict()
    expected_dict = {
        "tool": "prompt-issue",
        "name": "Prompt Issue Tool",
        "required": False,
        "schemaNodeId": None,
        "featureSchemaId": None,
        "classifications": [
            {
                "type": "checklist",
                "instructions": "prompt_issue",
                "name": "prompt_issue",
                "required": False,
                "options": [
                    {
                        "schemaNodeId": None,
                        "featureSchemaId": None,
                        "label": "This prompt cannot be rated (eg. contains PII, a nonsense prompt, a foreign language, or other scenario that makes the responses impossible to assess reliably). However, if you simply do not have expertise to tackle this prompt, please skip the task; do not mark it as not rateable.",
                        "value": "not_rateable",
                        "options": [],
                    },
                    {
                        "schemaNodeId": None,
                        "featureSchemaId": None,
                        "label": "This prompt contains a false, offensive, or controversial premise (eg. “why does 1+1=3”?)",
                        "value": "false_offensive_controversial",
                        "options": [],
                    },
                    {
                        "schemaNodeId": None,
                        "featureSchemaId": None,
                        "label": "This prompt is not self-contained, i.e. the prompt cannot be understood without additional context about previous turns, account information or images.",
                        "value": "not_self_contained",
                        "options": [],
                    },
                ],
                "schemaNodeId": None,
                "featureSchemaId": None,
                "scope": "global",
                "attributes": None,
            }
        ],
        "color": None,
    }
    assert tool_dict == expected_dict


def test_classification_validation():
    tool = PromptIssueTool(name="Prompt Issue Tool")
    with pytest.raises(ValueError):
        tool.classifications = [
            Classification(Classification.Type.TEXT, "prompt_issue")
        ]

    with pytest.raises(ValueError):
        tool.classifications = "abcd"

    # Test that we can modify the classification options
    classification = tool.classifications[0]
    classification.options.pop()
    classification.options[0].label = "test"
    classification.options[0].value == "test_value"
    tool.classifications = [classification]
    assert tool.classifications == [classification]



from io import BytesIO

import numpy as np
import pytest
from PIL import Image

from labelbox.data.annotation_types.data import GenericDataRowData, MaskData
from pydantic import ValidationError


def test_validate_schema():
    with pytest.raises(ValidationError):
        MaskData()


def test_im_bytes():
    data = (np.random.random((32, 32, 3)) * 255).astype(np.uint8)
    im_bytes = BytesIO()
    Image.fromarray(data).save(im_bytes, format="PNG")
    raster_data = MaskData(im_bytes=im_bytes.getvalue())
    data_ = raster_data.value
    assert np.all(data == data_)


def test_im_url():
    raster_data = MaskData(
        uid="test", url="https://picsum.photos/id/829/200/300"
    )
    data_ = raster_data.value
    assert data_.shape == (300, 200, 3)


def test_ref():
    external_id = "external_id"
    uid = "uid"
    metadata = []
    media_attributes = {}
    data = GenericDataRowData(
        uid=uid,
        metadata=metadata,
        media_attributes=media_attributes,
    )
    assert data.uid == uid
    assert data.media_attributes == media_attributes
    assert data.metadata == metadata



# type: ignore
from labelbox.data.annotation_types.metrics import ConfusionMatrixMetric
from typing import List, Union
from ...annotation_types import (
    ObjectAnnotation,
    ClassificationAnnotation,
)

from ..group import get_feature_pairs
from .calculation import confusion_matrix
from .calculation import feature_confusion_matrix


def confusion_matrix_metric(
    ground_truths: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    predictions: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    include_subclasses=False,
    iou=0.5,
) -> List[ConfusionMatrixMetric]:
    """
    Computes confusion matrix metrics between two sets of annotations.
    These annotations should relate to the same data (image/video).
    On the front end these will be displayed as precision, recall, and f1 scores.

    Args:
        ground_truth : Label containing human annotations or annotations known to be correct
        prediction: Label representing model predictions
        include_subclasses (bool): Whether or not to include subclasses in the calculation.
            If set to True, the iou between two overlapping objects of the same type is 0 if the subclasses are not the same.
    Returns:
        Returns a list of ConfusionMatrixMetrics. Will be empty if there were no predictions and labels. Otherwise a single metric will be returned.
    """
    if not (0.0 < iou < 1.0):
        raise ValueError("iou must be between 0 and 1")

    value = confusion_matrix(
        ground_truths, predictions, include_subclasses, iou
    )
    # If both gt and preds are empty there is no metric
    if value is None:
        return []

    metric_name = _get_metric_name(ground_truths, predictions, iou)
    return [ConfusionMatrixMetric(metric_name=metric_name, value=value)]


def feature_confusion_matrix_metric(
    ground_truths: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    predictions: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    include_subclasses=False,
    iou: float = 0.5,
) -> List[ConfusionMatrixMetric]:
    """
    Computes the confusion matrix metrics for each type of class in the list of annotations.
    These annotations should relate to the same data (image/video).
    On the front end these will be displayed as precision, recall, and f1 scores.

    Args:
        ground_truth : Label containing human annotations or annotations known to be correct
        prediction: Label representing model predictions
        include_subclasses (bool): Whether or not to include subclasses in the calculation.
            If set to True, the iou between two overlapping objects of the same type is 0 if the subclasses are not the same.
    Returns:
        Returns a list of ConfusionMatrixMetrics.
        There will be one metric for each class in the union of ground truth and prediction classes.
    """
    # Classifications are supported because we just take a naive approach to them..
    annotation_pairs = get_feature_pairs(ground_truths, predictions)
    metrics = []
    for key in annotation_pairs:
        value = feature_confusion_matrix(
            annotation_pairs[key][0],
            annotation_pairs[key][1],
            include_subclasses,
            iou,
        )
        if value is None:
            continue

        metric_name = _get_metric_name(
            annotation_pairs[key][0], annotation_pairs[key][1], iou
        )
        metrics.append(
            ConfusionMatrixMetric(
                metric_name=metric_name, feature_name=key, value=value
            )
        )
    return metrics


def _get_metric_name(
    ground_truths: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    predictions: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    iou: float,
):
    if _is_classification(ground_truths, predictions):
        return "classification"

    return f"{int(iou*100)}pct_iou"


def _is_classification(
    ground_truths: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    predictions: List[Union[ObjectAnnotation, ClassificationAnnotation]],
):
    # Check if either the prediction or label contains a classification annotation
    return (
        len(predictions)
        and isinstance(predictions[0], ClassificationAnnotation)
        or len(ground_truths)
        and isinstance(ground_truths[0], ClassificationAnnotation)
    )



import time
from labelbox import Project
from labelbox.schema.identifiables import GlobalKeys, UniqueIds


def test_get_task_queue(project: Project):
    task_queues = project.task_queues()
    assert len(task_queues) == 3
    review_queue = next(
        tq for tq in task_queues if tq.queue_type == "MANUAL_REVIEW_QUEUE"
    )
    assert review_queue


def test_get_overview_no_details(project: Project):
    po = project.get_overview()

    assert isinstance(po.to_label, int)
    assert isinstance(po.in_review, int)
    assert isinstance(po.in_rework, int)
    assert isinstance(po.skipped, int)
    assert isinstance(po.done, int)
    assert isinstance(po.issues, int)
    assert isinstance(po.labeled, int)
    assert isinstance(po.total_data_rows, int)


def test_get_overview_with_details(project: Project):
    po = project.get_overview(details=True)

    assert isinstance(po.to_label, int)
    assert isinstance(po.in_review["data"], list)
    assert isinstance(po.in_review["total"], int)
    assert isinstance(po.in_rework["data"], list)
    assert isinstance(po.in_rework["total"], int)
    assert isinstance(po.skipped, int)
    assert isinstance(po.done, int)
    assert isinstance(po.issues, int)
    assert isinstance(po.labeled, int)
    assert isinstance(po.total_data_rows, int)


def _validate_moved(project, queue_name, data_row_count):
    timeout_seconds = 30
    sleep_time = 2
    while True:
        task_queues = project.task_queues()
        review_queue = next(
            tq for tq in task_queues if tq.queue_type == queue_name
        )

        if review_queue.data_row_count == data_row_count:
            break

        if timeout_seconds <= 0:
            raise AssertionError(
                "Timed out expecting data_row_count of 1 in the review queue"
            )

        timeout_seconds -= sleep_time
        time.sleep(sleep_time)


def test_move_to_task(configured_batch_project_with_label):
    project, _, data_row, _ = configured_batch_project_with_label
    task_queues = project.task_queues()

    review_queue = next(
        tq for tq in task_queues if tq.queue_type == "MANUAL_REVIEW_QUEUE"
    )
    project.move_data_rows_to_task_queue(
        UniqueIds([data_row.uid]), review_queue.uid
    )
    _validate_moved(project, "MANUAL_REVIEW_QUEUE", 1)

    review_queue = next(
        tq for tq in task_queues if tq.queue_type == "MANUAL_REWORK_QUEUE"
    )
    project.move_data_rows_to_task_queue(
        GlobalKeys([data_row.global_key]), review_queue.uid
    )
    _validate_moved(project, "MANUAL_REWORK_QUEUE", 1)

    review_queue = next(
        tq for tq in task_queues if tq.queue_type == "MANUAL_REVIEW_QUEUE"
    )
    project.move_data_rows_to_task_queue(
        UniqueIds([data_row.uid]), review_queue.uid
    )
    _validate_moved(project, "MANUAL_REVIEW_QUEUE", 1)



from typing import List, Optional, TypedDict

from labelbox.schema.media_type import MediaType

EXPORT_LIMIT = 30


class DataRowParams(TypedDict):
    data_row_details: Optional[bool]
    metadata_fields: Optional[bool]
    attachments: Optional[bool]
    embeddings: Optional[bool]
    media_type_override: Optional[MediaType]


class ProjectExportParams(DataRowParams):
    project_details: Optional[bool]
    label_details: Optional[bool]
    performance_details: Optional[bool]
    interpolated_frames: Optional[bool]


class CatalogExportParams(DataRowParams):
    project_details: Optional[bool]
    label_details: Optional[bool]
    performance_details: Optional[bool]
    model_run_ids: Optional[List[str]]
    project_ids: Optional[List[str]]
    interpolated_frames: Optional[bool]
    all_projects: Optional[bool]
    all_model_runs: Optional[bool]
    predictions: Optional[bool]


class ModelRunExportParams(DataRowParams):
    predictions: Optional[bool]
    model_run_details: Optional[bool]


def _validate_array_length(array, max_length, array_name):
    if len(array) > max_length:
        raise ValueError(f"{array_name} cannot exceed {max_length} items")


def validate_catalog_export_params(params: CatalogExportParams):
    if "model_run_ids" in params and params["model_run_ids"] is not None:
        _validate_array_length(
            params["model_run_ids"], EXPORT_LIMIT, "model_run_ids"
        )

    if "project_ids" in params and params["project_ids"] is not None:
        _validate_array_length(
            params["project_ids"], EXPORT_LIMIT, "project_ids"
        )



import json
import collections.abc
from labelbox import DataRow
from labelbox.schema.data_row_metadata import DataRowMetadataField

INTEGRATION_SNAPSHOT_DIRECTORY = "tests/integration/snapshots"
TEXT_SCHEMA_ID = "cko8s9r5v0001h2dk9elqdidh"


def test_task_errors(dataset, image_url, snapshot):
    client = dataset.client
    task = dataset.create_data_rows(
        [
            {
                DataRow.row_data: image_url,
                DataRow.metadata_fields: [
                    DataRowMetadataField(
                        schema_id=TEXT_SCHEMA_ID, value="some msg"
                    ),
                    DataRowMetadataField(
                        schema_id=TEXT_SCHEMA_ID, value="some msg 2"
                    ),
                ],
            },
        ]
    )

    assert task in client.get_user().created_tasks()
    task.wait_till_done()

    assert len(task.failed_data_rows) == 1
    assert (
        "A schemaId can only be specified once per DataRow : [cko8s9r5v0001h2dk9elqdidh]"
        in task.failed_data_rows[0]["message"]
    )
    assert len(task.failed_data_rows[0]["failedDataRows"][0]["metadata"]) == 2

    dt = client.get_task_by_id(task.uid)
    assert dt.status == "COMPLETE"
    assert len(dt.errors) == 1
    assert dt.errors[0]["message"].startswith(
        "A schemaId can only be specified once per DataRow"
    )
    assert dt.result is None


def test_task_success_json(dataset, image_url, snapshot):
    client = dataset.client
    task = dataset.create_data_rows(
        [
            {
                DataRow.row_data: image_url,
            },
        ]
    )
    assert task in client.get_user().created_tasks()
    task.wait_till_done()
    assert task.status == "COMPLETE"
    assert task.errors is None
    assert task.result is not None
    assert isinstance(task.result, collections.abc.Sequence)
    assert task.result_url is not None
    assert isinstance(task.result_url, str)
    task_result = task.result[0]
    assert "id" in task_result and isinstance(task_result["id"], str)
    assert "row_data" in task_result and isinstance(
        task_result["row_data"], str
    )
    snapshot.snapshot_dir = INTEGRATION_SNAPSHOT_DIRECTORY
    task_result["id"] = "DUMMY_ID"
    task_result["row_data"] = "https://dummy.url"
    snapshot.assert_match(
        json.dumps(task_result), "test_task.test_task_success_json.json"
    )
    assert len(task.result)

    dt = client.get_task_by_id(task.uid)
    assert dt.status == "COMPLETE"
    assert len(dt.result) == 1
    assert dt.errors is None



from unittest.mock import MagicMock

from labelbox.schema.labeling_service_dashboard import LabelingServiceDashboard


def test_no_tasks_remaining_count():
    labeling_service_dashboard_data = {
        "id": "cm0eeo4c301lg07061phfhva0",
        "name": "TestStatus",
        "boostRequestedAt": "2024-08-28T22:08:07.446Z",
        "boostUpdatedAt": "2024-08-28T22:08:07.446Z",
        "boostRequestedBy": None,
        "boostStatus": "SET_UP",
        "dataRowsCount": 0,
        "dataRowsDoneCount": 0,
        "dataRowsInReviewCount": 0,
        "dataRowsInReworkCount": 0,
        "tasksTotalCount": 0,
        "tasksCompletedCount": 0,
        "tasksRemainingCount": 0,
        "mediaType": "image",
        "editorTaskType": None,
        "tags": [],
        "client": MagicMock(),
    }
    lsd = LabelingServiceDashboard(**labeling_service_dashboard_data)
    assert lsd.tasks_remaining_count is None


def test_tasks_remaining_count_exists():
    labeling_service_dashboard_data = {
        "id": "cm0eeo4c301lg07061phfhva0",
        "name": "TestStatus",
        "boostRequestedAt": "2024-08-28T22:08:07.446Z",
        "boostUpdatedAt": "2024-08-28T22:08:07.446Z",
        "boostRequestedBy": None,
        "boostStatus": "SET_UP",
        "dataRowsCount": 0,
        "dataRowsDoneCount": 0,
        "dataRowsInReviewCount": 0,
        "dataRowsInReworkCount": 0,
        "tasksTotalCount": 0,
        "tasksCompletedCount": 0,
        "tasksRemainingCount": 1,
        "mediaType": "image",
        "editorTaskType": None,
        "tags": [],
        "client": MagicMock(),
    }
    lsd = LabelingServiceDashboard(**labeling_service_dashboard_data)
    assert lsd.tasks_remaining_count == 1


def test_tasks_total_no_tasks_remaining_count():
    labeling_service_dashboard_data = {
        "id": "cm0eeo4c301lg07061phfhva0",
        "name": "TestStatus",
        "boostRequestedAt": "2024-08-28T22:08:07.446Z",
        "boostUpdatedAt": "2024-08-28T22:08:07.446Z",
        "boostRequestedBy": None,
        "boostStatus": "SET_UP",
        "dataRowsCount": 0,
        "dataRowsDoneCount": 0,
        "dataRowsInReviewCount": 1,
        "dataRowsInReworkCount": 0,
        "tasksTotalCount": 1,
        "tasksCompletedCount": 0,
        "tasksRemainingCount": 0,
        "mediaType": "image",
        "editorTaskType": None,
        "tags": [],
        "client": MagicMock(),
    }
    lsd = LabelingServiceDashboard(**labeling_service_dashboard_data)
    assert lsd.tasks_remaining_count == 0



import pytest
from lbox.exceptions import OperationNotSupportedException

from labelbox import LabelingFrontend


def test_get_labeling_frontends(client):
    filtered_frontends = list(
        client._get_labeling_frontends(where=LabelingFrontend.name == "Editor")
    )
    assert len(filtered_frontends)


def test_labeling_frontend_connecting_to_project(project):
    client = project.client
    default_labeling_frontend = next(
        client._get_labeling_frontends(where=LabelingFrontend.name == "Editor")
    )

    assert project.labeling_frontend() is None

    project.labeling_frontend.connect(default_labeling_frontend)
    assert project.labeling_frontend() == default_labeling_frontend

    with pytest.raises(OperationNotSupportedException):
        project.labeling_frontend.disconnect(default_labeling_frontend)



from typing import Optional
from pydantic import BaseModel, model_validator

from ...annotated_types import Cuid


class FeatureSchema(BaseModel):
    """
    Class that represents a feature schema.
    Could be a annotation, a subclass, or an option.
    Schema ids might not be known when constructing these objects so both a name and schema id are valid.
    """

    name: Optional[str] = None
    feature_schema_id: Optional[Cuid] = None

    @model_validator(mode="after")
    def must_set_one(self):
        if self.feature_schema_id is None and self.name is None:
            raise ValueError(
                "Must set either feature_schema_id or name for all feature schemas"
            )
        return self



from labelbox.schema.project import Project


def test_user(client):
    user = client.get_user()
    assert user.uid is not None
    assert user.organization() == client.get_organization()


def test_organization(client):
    organization = client.get_organization()
    assert organization.uid is not None
    assert client.get_user() in set(organization.users())


def test_user_and_org_projects(client, project):
    user = client.get_user()
    org = client.get_organization()
    user_project = user.projects(where=Project.uid == project.uid)
    org_project = org.projects(where=Project.uid == project.uid)

    assert user_project
    assert org_project



from labelbox.schema.identifiables import GlobalKeys, UniqueIds


def test_unique_ids():
    ids = ["a", "b", "c"]
    identifiables = UniqueIds(ids)
    assert [i for i in identifiables] == ids
    assert identifiables.id_type == "ID"
    assert len(identifiables) == 3


def test_global_keys():
    ids = ["a", "b", "c"]
    identifiables = GlobalKeys(ids)
    assert [i for i in identifiables] == ids
    assert identifiables.id_type == "GKEY"
    assert len(identifiables) == 3


def test_index_access():
    ids = ["a", "b", "c"]
    identifiables = GlobalKeys(ids)
    assert identifiables[0] == "a"
    assert identifiables[1:3] == GlobalKeys(["b", "c"])


def test_repr():
    ids = ["a", "b", "c"]
    identifiables = GlobalKeys(ids)
    assert repr(identifiables) == "GlobalKeys(['a', 'b', 'c'])"

    ids = ["a", "b", "c"]
    identifiables = UniqueIds(ids)
    assert repr(identifiables) == "UniqueIds(['a', 'b', 'c'])"



import logging
import threading
from queue import Queue
from typing import Any, Iterable
import threading

logger = logging.getLogger(__name__)


class ThreadSafeGen:
    """
    Wraps generators to make them thread safe
    """

    def __init__(self, iterable: Iterable[Any]):
        """ """
        self.iterable = iterable
        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return next(self.iterable)


class PrefetchGenerator:
    """
    Applys functions asynchronously to the output of a generator.
    Useful for modifying the generator results based on data from a network
    """

    def __init__(self, data: Iterable[Any], prefetch_limit=20, num_executors=1):
        if isinstance(data, (list, tuple)):
            self._data = (r for r in data)
        else:
            self._data = data

        self.queue = Queue(prefetch_limit)
        self.completed_threads = 0
        # Can only iterate over once it the queue.get hangs forever.
        self.multithread = num_executors > 1
        self.done = False

        if self.multithread:
            self._data = ThreadSafeGen(self._data)
            self.num_executors = num_executors
            self.threads = [
                threading.Thread(target=self.fill_queue)
                for _ in range(num_executors)
            ]
            for thread in self.threads:
                thread.daemon = True
                thread.start()
        else:
            self._data = iter(self._data)

    def _process(self, value) -> Any:
        raise NotImplementedError("Abstract method needs to be implemented")

    def fill_queue(self):
        try:
            for value in self._data:
                value = self._process(value)
                if value is None:
                    raise ValueError("Unexpected None")
                self.queue.put(value)
        except Exception as e:
            self.queue.put(
                ValueError(f"Unexpected exception while filling queue: {e}")
            )
        finally:
            self.queue.put(None)

    def __iter__(self):
        return self

    def __next__(self) -> Any:
        if self.done:
            raise StopIteration

        if self.multithread:
            value = self.queue.get()

            while value is None:
                self.completed_threads += 1
                if self.completed_threads == self.num_executors:
                    self.done = True
                    for thread in self.threads:
                        thread.join()
                    raise StopIteration
                value = self.queue.get()
            if isinstance(value, Exception):
                raise value
        else:
            value = self._process(next(self._data))
        return value



import itertools
from labelbox.schema.annotation_import import (
    AnnotationImport,
)
from labelbox.schema.media_type import MediaType
import pytest
from unittest.mock import patch


@patch("labelbox.schema.annotation_import.ANNOTATION_PER_LABEL_LIMIT", 1)
def test_above_annotation_limit_on_single_import_on_single_data_row(
    annotations_by_media_type,
):
    annotations_ndjson = list(
        itertools.chain.from_iterable(
            annotations_by_media_type[MediaType.Image]
        )
    )
    data_row_id = annotations_ndjson[0]["dataRow"]["id"]

    data_row_annotations = [
        annotation
        for annotation in annotations_ndjson
        if annotation["dataRow"]["id"] == data_row_id and "bbox" in annotation
    ]

    with pytest.raises(ValueError):
        AnnotationImport._validate_data_rows([data_row_annotations[0]] * 2)


@patch("labelbox.schema.annotation_import.ANNOTATION_PER_LABEL_LIMIT", 1)
def test_above_annotation_limit_divided_among_different_rows(
    annotations_by_media_type,
):
    annotations_ndjson = list(
        itertools.chain.from_iterable(
            annotations_by_media_type[MediaType.Image]
        )
    )
    data_row_id = annotations_ndjson[0]["dataRow"]["id"]

    first_data_row_annotation = [
        annotation
        for annotation in annotations_ndjson
        if annotation["dataRow"]["id"] == data_row_id and "bbox" in annotation
    ][0]

    second_data_row_annotation = first_data_row_annotation.copy()
    second_data_row_annotation["dataRow"]["id"] == "data_row_id_2"

    with pytest.raises(ValueError):
        AnnotationImport._validate_data_rows(
            [first_data_row_annotation, second_data_row_annotation] * 2
        )



from labelbox.utils import _CamelCaseMixin


from datetime import datetime
from typing import Dict
from pydantic import BaseModel


class Model(_CamelCaseMixin, BaseModel):
    id: str
    description: str
    inference_params_json_schema: Dict
    name: str
    ontology_id: str
    created_at: datetime


MODEL_FIELD_NAMES = list(Model.model_json_schema()["properties"].keys())



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from labelbox.schema.tool_building.classification import (
    Classification,
    Option,
)
from labelbox.schema.tool_building.tool_type import ToolType


def _supported_classifications() -> List[Classification]:
    option_1_text = "This prompt cannot be rated (eg. contains PII, a nonsense prompt, a foreign language, or other scenario that makes the responses impossible to assess reliably). However, if you simply do not have expertise to tackle this prompt, please skip the task; do not mark it as not rateable."
    option_2_text = "This prompt contains a false, offensive, or controversial premise (eg. “why does 1+1=3”?)"
    option_3_text = "This prompt is not self-contained, i.e. the prompt cannot be understood without additional context about previous turns, account information or images."
    options = [
        Option(label=option_1_text, value="not_rateable"),
        Option(label=option_2_text, value="false_offensive_controversial"),
        Option(label=option_3_text, value="not_self_contained"),
    ]
    return [
        Classification(
            class_type=Classification.Type.CHECKLIST,
            name="prompt_issue",
            options=options,
        ),
    ]


@dataclass
class PromptIssueTool:
    """
    Use this class in OntologyBuilder to create a tool for prompt rating
    It comes with a prebuild checklist of options, which a user can modify or override
    So essentially this is a tool with a prebuilt checklist classification
    """

    name: str
    type: ToolType = field(default=ToolType.PROMPT_ISSUE, init=False)
    required: bool = False  # This attribute is for consistency with other tools and backend, default is False
    schema_id: Optional[str] = None
    feature_schema_id: Optional[str] = None
    color: Optional[str] = None
    classifications: List[Classification] = field(
        default_factory=_supported_classifications
    )

    def __post_init__(self):
        if self.name.strip() == "":
            raise ValueError("Name cannot be empty")

        if not self._validate_classifications(self.classifications):
            raise ValueError("Only one checklist classification is supported")

    def __setattr__(self, name, value):
        if name == "classifications" and not self._validate_classifications(
            value
        ):
            raise ValueError("Classifications are immutable")
        object.__setattr__(self, name, value)

    def _validate_classifications(
        self, classifications: List[Classification]
    ) -> bool:
        if (
            len(classifications) != 1
            or classifications[0].class_type != Classification.Type.CHECKLIST
            or len(classifications[0].options) < 1
        ):
            return False
        return True

    def asdict(self) -> Dict[str, Any]:
        classifications_valid = self._validate_classifications(
            self.classifications
        )
        if not classifications_valid:
            raise ValueError(
                "Classifications for Prompt Issue Tool are invalid"
            )

        return {
            "tool": self.type.value,
            "name": self.name,
            "required": self.required,
            "schemaNodeId": self.schema_id,
            "featureSchemaId": self.feature_schema_id,
            "classifications": [
                classification.asdict()
                for classification in self.classifications
            ],
            "color": self.color,
        }

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> "PromptIssueTool":
        return cls(
            name=dictionary["name"],
            schema_id=dictionary.get("schemaNodeId", None),
            feature_schema_id=dictionary.get("featureSchemaId", None),
            required=dictionary.get("required", False),
            classifications=[
                Classification.from_dict(classification)
                for classification in dictionary["classifications"]
            ],
            color=dictionary.get("color", None),
        )



import json

from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
from labelbox.data.annotation_types.metrics.confusion_matrix import (
    ConfusionMatrixMetric,
)
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.types import (
    Label,
    ScalarMetric,
    ScalarMetricAggregation,
    ConfusionMatrixAggregation,
)


def test_metric():
    with open("tests/data/assets/ndjson/metric_import.json", "r") as file:
        data = json.load(file)

    labels = [
        Label(
            data=GenericDataRowData(
                uid="ckrmdnqj4000007msh9p2a27r",
            ),
            annotations=[
                ScalarMetric(
                    value=0.1,
                    extra={"uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7672"},
                    aggregation=ScalarMetricAggregation.ARITHMETIC_MEAN,
                )
            ],
        )
    ]

    res = list(NDJsonConverter.serialize(labels))
    assert res == data


def test_custom_scalar_metric():
    data = [
        {
            "uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7672",
            "dataRow": {"id": "ckrmdnqj4000007msh9p2a27r"},
            "metricValue": 0.1,
            "metricName": "custom_iou",
            "featureName": "sample_class",
            "subclassName": "sample_subclass",
            "aggregation": "SUM",
        },
        {
            "uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7673",
            "dataRow": {"id": "ckrmdnqj4000007msh9p2a27r"},
            "metricValue": 0.1,
            "metricName": "custom_iou",
            "featureName": "sample_class",
            "aggregation": "SUM",
        },
        {
            "uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7674",
            "dataRow": {"id": "ckrmdnqj4000007msh9p2a27r"},
            "metricValue": {0.1: 0.1, 0.2: 0.5},
            "metricName": "custom_iou",
            "aggregation": "SUM",
        },
    ]

    labels = [
        Label(
            data=GenericDataRowData(
                uid="ckrmdnqj4000007msh9p2a27r",
            ),
            annotations=[
                ScalarMetric(
                    value=0.1,
                    feature_name="sample_class",
                    subclass_name="sample_subclass",
                    extra={"uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7672"},
                    metric_name="custom_iou",
                    aggregation=ScalarMetricAggregation.SUM,
                ),
                ScalarMetric(
                    value=0.1,
                    feature_name="sample_class",
                    extra={"uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7673"},
                    metric_name="custom_iou",
                    aggregation=ScalarMetricAggregation.SUM,
                ),
                ScalarMetric(
                    value={"0.1": 0.1, "0.2": 0.5},
                    extra={"uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7674"},
                    metric_name="custom_iou",
                    aggregation=ScalarMetricAggregation.SUM,
                ),
            ],
        )
    ]

    res = list(NDJsonConverter.serialize(labels))

    assert res == data


def test_custom_confusion_matrix_metric():
    data = [
        {
            "uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7672",
            "dataRow": {"id": "ckrmdnqj4000007msh9p2a27r"},
            "metricValue": (1, 1, 2, 3),
            "metricName": "50%_iou",
            "featureName": "sample_class",
            "subclassName": "sample_subclass",
            "aggregation": "CONFUSION_MATRIX",
        },
        {
            "uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7673",
            "dataRow": {"id": "ckrmdnqj4000007msh9p2a27r"},
            "metricValue": (0, 1, 2, 5),
            "metricName": "50%_iou",
            "featureName": "sample_class",
            "aggregation": "CONFUSION_MATRIX",
        },
        {
            "uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7674",
            "dataRow": {"id": "ckrmdnqj4000007msh9p2a27r"},
            "metricValue": {0.1: (0, 1, 2, 3), 0.2: (5, 3, 4, 3)},
            "metricName": "50%_iou",
            "aggregation": "CONFUSION_MATRIX",
        },
    ]

    labels = [
        Label(
            data=GenericDataRowData(
                uid="ckrmdnqj4000007msh9p2a27r",
            ),
            annotations=[
                ConfusionMatrixMetric(
                    value=(1, 1, 2, 3),
                    feature_name="sample_class",
                    subclass_name="sample_subclass",
                    extra={"uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7672"},
                    metric_name="50%_iou",
                    aggregation=ConfusionMatrixAggregation.CONFUSION_MATRIX,
                ),
                ConfusionMatrixMetric(
                    value=(0, 1, 2, 5),
                    feature_name="sample_class",
                    extra={"uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7673"},
                    metric_name="50%_iou",
                    aggregation=ConfusionMatrixAggregation.CONFUSION_MATRIX,
                ),
                ConfusionMatrixMetric(
                    value={0.1: (0, 1, 2, 3), 0.2: (5, 3, 4, 3)},
                    extra={"uuid": "a22bbf6e-b2da-4abe-9a11-df84759f7674"},
                    metric_name="50%_iou",
                    aggregation=ConfusionMatrixAggregation.CONFUSION_MATRIX,
                ),
            ],
        )
    ]

    res = list(NDJsonConverter.serialize(labels))

    assert data == res



from pytest_cases import fixture_ref
from pytest_cases import parametrize, fixture_ref

from labelbox.data.metrics.confusion_matrix.confusion_matrix import (
    feature_confusion_matrix_metric,
)


@parametrize(
    "tool_examples",
    [
        fixture_ref("polygon_pairs"),
        fixture_ref("rectangle_pairs"),
        fixture_ref("mask_pairs"),
        fixture_ref("line_pairs"),
        fixture_ref("point_pairs"),
        fixture_ref("ner_pairs"),
    ],
)
def test_overlapping_objects(tool_examples):
    for example in tool_examples:
        for include_subclasses, expected_attr_name in [
            [True, "expected"],
            [False, "expected_without_subclasses"],
        ]:
            metrics = feature_confusion_matrix_metric(
                example.ground_truths,
                example.predictions,
                include_subclasses=include_subclasses,
            )

            metrics = {r.feature_name: list(r.value) for r in metrics}
            if len(getattr(example, expected_attr_name)) == 0:
                assert len(metrics) == 0
            else:
                assert metrics == getattr(
                    example, expected_attr_name
                ), f"{example.predictions},{example.ground_truths}"


@parametrize(
    "tool_examples",
    [fixture_ref("checklist_pairs"), fixture_ref("radio_pairs")],
)
def test_overlapping_classifications(tool_examples):
    for example in tool_examples:
        metrics = feature_confusion_matrix_metric(
            example.ground_truths, example.predictions
        )

        metrics = {r.feature_name: list(r.value) for r in metrics}
        if len(example.expected) == 0:
            assert len(metrics) == 0
        else:
            assert (
                metrics == example.expected
            ), f"{example.predictions},{example.ground_truths}"



from typing import TYPE_CHECKING

from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field, Relationship

if TYPE_CHECKING:
    from labelbox import Project, Role


class User(DbObject):
    """A User is a registered Labelbox user (for example you) associated with
    data they create or import and an Organization they belong to.

    Attributes:
        updated_at (datetime)
        created_at (datetime)
        email (str)
        name (str)
        nickname (str)
        picture (str)
        is_viewer (bool)
        is_external_viewer (bool)

        organization (Relationship): `ToOne` relationship to Organization
        created_tasks (Relationship): `ToMany` relationship to Task
        projects (Relationship): `ToMany` relationship to Project
    """

    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    email = Field.String("email")
    name = Field.String("nickname")
    nickname = Field.String("name")
    picture = Field.String("picture")
    is_viewer = Field.Boolean("is_viewer")
    is_external_user = Field.Boolean("is_external_user")

    # Relationships
    organization = Relationship.ToOne("Organization")
    created_tasks = Relationship.ToMany("Task", False, "created_tasks")
    projects = Relationship.ToMany("Project", False)
    org_role = Relationship.ToOne("OrgRole", False)

    def update_org_role(self, role: "Role") -> None:
        """Updated the `User`s organization role.

        See client.get_roles() to get all valid roles
        If you a user is converted from project level permissions to org level permissions and then convert back, their permissions will remain for each individual project

        Args:
            role (Role): The role that you want to set for this user.

        """
        user_id_param = "userId"
        role_id_param = "roleId"
        query_str = """mutation SetOrganizationRolePyApi($%s: ID!, $%s: ID!) {
            setOrganizationRole(data: {userId: $userId, roleId: $roleId}) { id name }}
        """ % (user_id_param, role_id_param)

        self.client.execute(
            query_str, {user_id_param: self.uid, role_id_param: role.uid}
        )

    def remove_from_project(self, project: "Project") -> None:
        """Removes a User from a project. Only used for project based users.
        Project based user means their org role is "NONE"

        Args:
            project (Project): Project to remove user from

        """
        self.upsert_project_role(project, self.client.get_roles()["NONE"])

    def upsert_project_role(self, project: "Project", role: "Role") -> None:
        """Updates or replaces a User's role in a project.

        Args:
            project (Project): The project to update the users permissions for
            role (Role): The role to assign to this user in this project.

        """
        org_role = self.org_role()
        if org_role.name.upper() != "NONE":
            raise ValueError(
                "User is not project based and has access to all projects"
            )

        project_id_param = "projectId"
        user_id_param = "userId"
        role_id_param = "roleId"
        query_str = """mutation SetProjectMembershipPyApi($%s: ID!, $%s: ID!, $%s: ID!) {
                setProjectMembership(data: {%s: $userId, roleId: $%s, projectId: $%s}) {id}}
        """ % (
            user_id_param,
            role_id_param,
            project_id_param,
            user_id_param,
            role_id_param,
            project_id_param,
        )

        self.client.execute(
            query_str,
            {
                project_id_param: project.uid,
                user_id_param: self.uid,
                role_id_param: role.uid,
            },
        )
